/*
 * Copyright (C) 2005-2020 Darron Broad
 * All rights reserved.
 *
 * This file is part of Pickle Microchip PIC ICSP.
 *
 * Pickle Microchip PIC ICSP is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * Pickle Microchip PIC ICSP is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Pickle Microchip PIC ICSP. If not, see http://www.gnu.org/licenses/
 */

#undef DEBUG

#include "pickle.h"

/******************************************************************************
 *
 * Session
 *
 *****************************************************************************/

extern struct pickle p;

/*****************************************************************************
 *
 * Hardware operations
 *
 *****************************************************************************/

struct pic_ops pic24_ops = {
	.arch				= ARCH24BIT,
	.align				= sizeof(uint32_t),
	.selector			= pic24_selector,
#ifdef LOADER
	.bootloader			= pic24_bootloader,
#else
	.bootloader			= NULL,
#endif
	.program_begin			= pic24_program_begin,
	.program_data			= pic24_program_data,
	.program_end			= pic24_program_end,
	.verify_begin			= pic24_program_verify,
	.verify_data			= pic24_verify_data,
	.verify_end			= pic24_standby,
	.view_data			= pic24_view_data,
	.read_config_memory		= pic24_read_config_memory,
	.get_program_count		= pic24_get_program_count,
	.get_program_size		= pic24_get_program_size,
	.get_data_size			= pic24_get_data_size,
	.get_executive_size		= pic24_get_executive_size,
	.get_boot_size			= NULL,
	.read_program_memory_block	= pic24_read_program_memory_block,
	.read_data_memory_block		= pic24_read_data_memory_block,
	.write_panel			= pic24_write_panel,
	.bulk_erase			= pic24_bulk_erase,
	.write_osccal			= NULL,
	.write_bandgap			= NULL,
	.write_calib			= NULL,
	.row_erase			= pic24_row_erase,
	.dumpdeviceid			= pic24_dumpdeviceid,
	.dumpconfig			= pic24_dumpconfig,
	.dumposccal			= NULL,
	.dumpdevice			= pic24_dumpdevice,
	.dumpadj			= 2,
	.dumphexcode			= pic24_dumphexcode,
	.dumpinhxcode			= pic24_dumpinhxcode,
	.dumphexdata			= pic24_dumphexdata,
	.dumpinhxdata			= pic24_dumpinhxdata,
	.debug				= NULL,
};

uint32_t
pic24_arch(void)
{
	p.pic = &pic24_ops;

	return p.pic->arch;
}

/*****************************************************************************
 *
 * Hardware configuration
 *
 *****************************************************************************/

struct pic24_config pic24_conf;

/*****************************************************************************
 *
 * Hardware algorithm map: Device
 *
 *****************************************************************************/

struct pic24_dsmap pic24_map[] =
{
/*Device name          Device id           Flash  EEPROM Data-sheet Config    #   Fuid #      Calib #     PanelSizes BOOT/Config/MEM*/
/* These devices have 2 config words at the end of program flash */
{"PIC24FJ16GA002",     PIC24FJ16GA002,     5632,  0,     DS39768D,  0x002BFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ16GA004",     PIC24FJ16GA004,     5632,  0,     DS39768D,  0x002BFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ32GA002",     PIC24FJ32GA002,     11264, 0,     DS39768D,  0x0057FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ32GA004",     PIC24FJ32GA004,     11264, 0,     DS39768D,  0x0057FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ48GA002",     PIC24FJ48GA002,     16896, 0,     DS39768D,  0x0083FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ48GA004",     PIC24FJ48GA004,     16896, 0,     DS39768D,  0x0083FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA002",     PIC24FJ64GA002,     22016, 0,     DS39768D,  0x00ABFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA004",     PIC24FJ64GA004,     22016, 0,     DS39768D,  0x00ABFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA006",     PIC24FJ64GA006,     22016, 0,     DS39768D,  0x00ABFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA008",     PIC24FJ64GA008,     22016, 0,     DS39768D,  0x00ABFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA010",     PIC24FJ64GA010,     22016, 0,     DS39768D,  0x00ABFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ96GA006",     PIC24FJ96GA006,     32768, 0,     DS39768D,  0x00FFFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ96GA008",     PIC24FJ96GA008,     32768, 0,     DS39768D,  0x00FFFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ96GA010",     PIC24FJ96GA010,     32768, 0,     DS39768D,  0x00FFFC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA006",    PIC24FJ128GA006,    44032, 0,     DS39768D,  0x0157FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA008",    PIC24FJ128GA008,    44032, 0,     DS39768D,  0x0157FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA010",    PIC24FJ128GA010,    44032, 0,     DS39768D,  0x0157FC, 2, 0,0,         0x8007F4, 6, 64,0,512,0,   0,0,0},

/* These devices have 7 config words in the config space */
{"dsPIC30F2010",       dsPIC30F2010,       4096,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F2011",       dsPIC30F2011,       4096,  0,     DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F2012",       dsPIC30F2012,       4096,  0,     DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F3010",       dsPIC30F3010,       8192,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F3011",       dsPIC30F3011,       8192,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F3012",       dsPIC30F3012,       8192,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F3013",       dsPIC30F3013,       8192,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F3014",       dsPIC30F3014,       8192,  1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F4011",       dsPIC30F4011,       16384, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F4012",       dsPIC30F4012,       16384, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F4013",       dsPIC30F4013,       16384, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F5011",       dsPIC30F5011,       22528, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F5013",       dsPIC30F5013,       22528, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F5015",       dsPIC30F5015,       22528, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F5016",       dsPIC30F5016,       22528, 1024,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6010",       dsPIC30F6010,       49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6010A",      dsPIC30F6010A,      49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6011",       dsPIC30F6011,       45056, 2048,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6011A",      dsPIC30F6011A,      45056, 2048,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6012",       dsPIC30F6012,       49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6012A",      dsPIC30F6012A,      49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6013",       dsPIC30F6013,       45056, 2048,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6013A",      dsPIC30F6013A,      45056, 2048,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6014",       dsPIC30F6014,       49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6014A",      dsPIC30F6014A,      49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},
{"dsPIC30F6015",       dsPIC30F6015,       49152, 4096,  DS70102H,  0xF80000, 7, 0x8005C0,32, 0, 0,        32,16,32,0,   0,0,0},

/* These devices have 2 config words at the end of program flash */
{"dsPIC33FJ16GP101",   dsPIC33FJ16GP101,   5632,  0,     DS70659C,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ16GP102",   dsPIC33FJ16GP102,   5632,  0,     DS70659C,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32GP101",   dsPIC33FJ32GP101,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32GP102",   dsPIC33FJ32GP102,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32GP104",   dsPIC33FJ32GP104,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ16MC101",   dsPIC33FJ16MC101,   5632,  0,     DS70659C,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ16MC102",   dsPIC33FJ16MC102,   5632,  0,     DS70659C,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32MC101",   dsPIC33FJ32MC101,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32MC102",   dsPIC33FJ32MC102,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ32MC104",   dsPIC33FJ32MC104,   11264, 0,     DS70659C,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},

/* These devices have 8 config bytes at the end of program flash */
{"dsPIC33FJ06GS001",   dsPIC33FJ06GS001,   2048,  0,     DS70659C,  0x000FF0, 8, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ06GS101A",  dsPIC33FJ06GS101A,  2048,  0,     DS70659C,  0x000FF0, 8, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ06GS102A",  dsPIC33FJ06GS102A,  2048,  0,     DS70659C,  0x000FF0, 8, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ06GS202A",  dsPIC33FJ06GS202A,  2048,  0,     DS70659C,  0x000FF0, 8, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"dsPIC33FJ09GS302",   dsPIC33FJ09GS302,   3072,  0,     DS70659C,  0x0017F0, 8, 0, 0,        0, 0,        1,0,512,0,    0,0,0},

{"PIC24F08KA101",      PIC24F08KA101,      2816,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KA101",      PIC24F16KA101,      5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KA102",      PIC24F08KA102,      2816,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KA102",      PIC24F16KA102,      5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KA301",     PIC24FV16KA301,     5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KA302",     PIC24FV16KA302,     5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KA304",     PIC24FV16KA304,     5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV32KA301",     PIC24FV32KA301,     11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV32KA302",     PIC24FV32KA302,     11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV32KA304",     PIC24FV32KA304,     11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KA301",      PIC24F16KA301,      5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KA302",      PIC24F16KA302,      5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KA304",      PIC24F16KA304,      5632,  512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F32KA301",      PIC24F32KA301,      11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F32KA302",      PIC24F32KA302,      11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F32KA304",      PIC24F32KA304,      11264, 512,   DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F04KA200",      PIC24F04KA200,      1408,  0,     DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},
{"PIC24F04KA201",      PIC24F04KA201,      1408,  0,     DS39919C,  0xF80000, 9, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},

{"PIC24F16KL402",      PIC24F16KL402,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KL401",      PIC24F16KL401,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KL402",      PIC24F08KL402,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KL401",      PIC24F08KL401,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KL302",      PIC24F08KL302,      2816,  256,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KL301",      PIC24F08KL301,      2816,  256,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KL201",      PIC24F08KL201,      2816,  0,     DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},
{"PIC24F08KL200",      PIC24F08KL200,      2816,  0,     DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},
{"PIC24F04KL101",      PIC24F04KL101,      1408,  0,     DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},
{"PIC24F04KL100",      PIC24F04KL100,      1408,  0,     DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,0,32,0,    0,0,0},
{"PIC24FV16KM204",     PIC24FV16KM204,     5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV08KM204",     PIC24FV08KM204,     2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KM104",     PIC24FV16KM104,     5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KM202",     PIC24FV16KM202,     5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV08KM202",     PIC24FV08KM202,     2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV16KM102",     PIC24FV16KM102,     5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV08KM102",     PIC24FV08KM102,     2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24FV08KM101",     PIC24FV08KM101,     2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KM204",      PIC24F16KM204,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KM204",      PIC24F08KM204,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KM104",      PIC24F16KM104,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KM202",      PIC24F16KM202,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KM202",      PIC24F08KM202,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F16KM102",      PIC24F16KM102,      5632,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KM102",      PIC24F08KM102,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},
{"PIC24F08KM101",      PIC24F08KM101,      2816,  512,   DS30625D,  0xF80000, 8, 0, 0,        0x8007F4, 6, 32,1,32,0,    0,0,0},

{"dsPIC33FJ06GS101",   dsPIC33FJ06GS101,   2048,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ06GS102",   dsPIC33FJ06GS102,   2048,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ06GS202",   dsPIC33FJ06GS202,   2048,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ16GS402",   dsPIC33FJ16GS402,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ16GS404",   dsPIC33FJ16GS404,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ16GS502",   dsPIC33FJ16GS502,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ16GS504",   dsPIC33FJ16GS504,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 2, 0x8007F4, 6, 64,0,512,0,   0,0,0},

{"dsPIC33FJ32GS406",   dsPIC33FJ32GS406,   11264, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ32GS606",   dsPIC33FJ32GS606,   11264, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ32GS608",   dsPIC33FJ32GS608,   11264, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ32GS610",   dsPIC33FJ32GS610,   11264, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ64GS406",   dsPIC33FJ64GS406,   22016, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ64GS606",   dsPIC33FJ64GS606,   22016, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ64GS608",   dsPIC33FJ64GS608,   22016, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},
{"dsPIC33FJ64GS610",   dsPIC33FJ64GS610,   22016, 0,     DS70152H,  0xF80000, 9, 0, 0,        0x8007F4, 6, 64,0,512,0,   0,0,0},

{"dsPIC33FJ12GP201",   dsPIC33FJ12GP201,   4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ12GP202",   dsPIC33FJ12GP202,   4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ16GP304",   dsPIC33FJ16GP304,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32GP202",   dsPIC33FJ32GP202,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32GP204",   dsPIC33FJ32GP204,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32GP302",   dsPIC33FJ32GP302,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32GP304",   dsPIC33FJ32GP304,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP202",   dsPIC33FJ64GP202,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP204",   dsPIC33FJ64GP204,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP206",   dsPIC33FJ64GP206,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP306",   dsPIC33FJ64GP306,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP310",   dsPIC33FJ64GP310,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP706",   dsPIC33FJ64GP706,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP708",   dsPIC33FJ64GP708,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP710",   dsPIC33FJ64GP710,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP802",   dsPIC33FJ64GP802,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP804",   dsPIC33FJ64GP804,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP202",  dsPIC33FJ128GP202,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP204",  dsPIC33FJ128GP204,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP206",  dsPIC33FJ128GP206,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP306",  dsPIC33FJ128GP306,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP310",  dsPIC33FJ128GP310,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP706",  dsPIC33FJ128GP706,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP708",  dsPIC33FJ128GP708,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP710",  dsPIC33FJ128GP710,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP802",  dsPIC33FJ128GP802,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP804",  dsPIC33FJ128GP804,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP506",  dsPIC33FJ256GP506,  87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP510",  dsPIC33FJ256GP510,  87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP710",  dsPIC33FJ256GP710,  87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ12MC201",   dsPIC33FJ12MC201,   4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ12MC202",   dsPIC33FJ12MC202,   4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ16MC304",   dsPIC33FJ16MC304,   5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32MC202",   dsPIC33FJ32MC202,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32MC204",   dsPIC33FJ32MC204,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32MC302",   dsPIC33FJ32MC302,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ32MC304",   dsPIC33FJ32MC304,   11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC202",   dsPIC33FJ64MC202,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC204",   dsPIC33FJ64MC204,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC506",   dsPIC33FJ64MC506,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC508",   dsPIC33FJ64MC508,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC510",   dsPIC33FJ64MC510,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC706",   dsPIC33FJ64MC706,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC710",   dsPIC33FJ64MC710,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC802",   dsPIC33FJ64MC802,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC804",   dsPIC33FJ64MC804,   22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC202",  dsPIC33FJ128MC202,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC204",  dsPIC33FJ128MC204,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC506",  dsPIC33FJ128MC506,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC510",  dsPIC33FJ128MC510,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC706",  dsPIC33FJ128MC706,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC708",  dsPIC33FJ128MC708,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC710",  dsPIC33FJ128MC710,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC802",  dsPIC33FJ128MC802,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC804",  dsPIC33FJ128MC804,  44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256MC510",  dsPIC33FJ256MC510,  87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256MC710",  dsPIC33FJ256MC710,  87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ12GP201",     PIC24HJ12GP201,     4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ12GP202",     PIC24HJ12GP202,     4096,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ16GP304",     PIC24HJ16GP304,     5632,  0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ32GP202",     PIC24HJ32GP202,     11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ32GP204",     PIC24HJ32GP204,     11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ32GP302",     PIC24HJ32GP302,     11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ32GP304",     PIC24HJ32GP304,     11264, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP202",     PIC24HJ64GP202,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP204",     PIC24HJ64GP204,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP206",     PIC24HJ64GP206,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP210",     PIC24HJ64GP210,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP502",     PIC24HJ64GP502,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP504",     PIC24HJ64GP504,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP506",     PIC24HJ64GP506,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP510",     PIC24HJ64GP510,     22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP202",    PIC24HJ128GP202,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP204",    PIC24HJ128GP204,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP206",    PIC24HJ128GP206,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP210",    PIC24HJ128GP210,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP306",    PIC24HJ128GP306,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP310",    PIC24HJ128GP310,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP502",    PIC24HJ128GP502,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP504",    PIC24HJ128GP504,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP506",    PIC24HJ128GP506,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP510",    PIC24HJ128GP510,    44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP206",    PIC24HJ256GP206,    87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP210",    PIC24HJ256GP210,    87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP610",    PIC24HJ256GP610,    87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP206A",  dsPIC33FJ64GP206A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP306A",  dsPIC33FJ64GP306A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP310A",  dsPIC33FJ64GP310A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP706A",  dsPIC33FJ64GP706A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP708A",  dsPIC33FJ64GP708A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64GP710A",  dsPIC33FJ64GP710A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC506A",  dsPIC33FJ64MC506A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC508A",  dsPIC33FJ64MC508A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC510A",  dsPIC33FJ64MC510A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC706A",  dsPIC33FJ64MC706A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ64MC710A",  dsPIC33FJ64MC710A,  22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP206A",    PIC24HJ64GP206A,    22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP210A",    PIC24HJ64GP210A,    22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP506A",    PIC24HJ64GP506A,    22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ64GP510A",    PIC24HJ64GP510A,    22016, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP206A", dsPIC33FJ128GP206A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP306A", dsPIC33FJ128GP306A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP310A", dsPIC33FJ128GP310A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP706A", dsPIC33FJ128GP706A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP708A", dsPIC33FJ128GP708A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128GP710A", dsPIC33FJ128GP710A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC506A", dsPIC33FJ128MC506A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC510A", dsPIC33FJ128MC510A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC706A", dsPIC33FJ128MC706A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC708A", dsPIC33FJ128MC708A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ128MC710A", dsPIC33FJ128MC710A, 44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP206A",   PIC24HJ128GP206A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP210A",   PIC24HJ128GP210A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP306A",   PIC24HJ128GP306A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP310A",   PIC24HJ128GP310A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP506A",   PIC24HJ128GP506A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ128GP510A",   PIC24HJ128GP510A,   44032, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP506A", dsPIC33FJ256GP506A, 87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP510A", dsPIC33FJ256GP510A, 87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256GP710A", dsPIC33FJ256GP710A, 87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256MC510A", dsPIC33FJ256MC510A, 87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"dsPIC33FJ256MC710A", dsPIC33FJ256MC710A, 87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP206A",   PIC24HJ256GP206A,   87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP210A",   PIC24HJ256GP210A,   87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},
{"PIC24HJ256GP610A",   PIC24HJ256GP610A,   87552, 0,     DS70152H,  0xF80000, 8, 0xF80010, 4, 0, 0,        64,0,512,0,   0,0,0},

/* These devices have 10 config bytes at the end of program flash */
{"PIC24EP32GP202",     PIC24EP32GP202,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP32GP203",     PIC24EP32GP203,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP32GP204",     PIC24EP32GP204,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32GP502",   dsPIC33EP32GP502,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32GP503",   dsPIC33EP32GP503,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32GP504",   dsPIC33EP32GP504,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP32MC202",     PIC24EP32MC202,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP32MC203",     PIC24EP32MC203,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP32MC204",     PIC24EP32MC204,     11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC202",   dsPIC33EP32MC202,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC203",   dsPIC33EP32MC203,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC204",   dsPIC33EP32MC204,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC502",   dsPIC33EP32MC502,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC503",   dsPIC33EP32MC503,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP32MC504",   dsPIC33EP32MC504,   11264, 0,     DS70663D,  0x0057EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"PIC24EP64GP202",     PIC24EP64GP202,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64GP203",     PIC24EP64GP203,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64GP204",     PIC24EP64GP204,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64GP206",     PIC24EP64GP206,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64GP502",   dsPIC33EP64GP502,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64GP503",   dsPIC33EP64GP503,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64GP504",   dsPIC33EP64GP504,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64GP506",   dsPIC33EP64GP506,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64MC202",     PIC24EP64MC202,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64MC203",     PIC24EP64MC203,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64MC204",     PIC24EP64MC204,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP64MC206",     PIC24EP64MC206,     22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC202",   dsPIC33EP64MC202,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC203",   dsPIC33EP64MC203,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC204",   dsPIC33EP64MC204,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC206",   dsPIC33EP64MC206,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC502",   dsPIC33EP64MC502,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC503",   dsPIC33EP64MC503,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC504",   dsPIC33EP64MC504,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP64MC506",   dsPIC33EP64MC506,   22528, 0,     DS70663D,  0x00AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128GP202",    PIC24EP128GP202,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128GP204",    PIC24EP128GP204,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128GP206",    PIC24EP128GP206,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128GP502",  dsPIC33EP128GP502,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128GP504",  dsPIC33EP128GP504,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128GP506",  dsPIC33EP128GP506,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128MC202",    PIC24EP128MC202,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128MC204",    PIC24EP128MC204,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP128MC206",    PIC24EP128MC206,    44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC202",  dsPIC33EP128MC202,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC204",  dsPIC33EP128MC204,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC206",  dsPIC33EP128MC206,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC502",  dsPIC33EP128MC502,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC504",  dsPIC33EP128MC504,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP128MC506",  dsPIC33EP128MC506,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256GP202",    PIC24EP256GP202,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256GP204",    PIC24EP256GP204,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256GP206",    PIC24EP256GP206,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256GP502",  dsPIC33EP256GP502,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256GP504",  dsPIC33EP256GP504,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256GP506",  dsPIC33EP256GP506,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256MC202",    PIC24EP256MC202,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256MC204",    PIC24EP256MC204,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP256MC206",    PIC24EP256MC206,    88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC202",  dsPIC33EP256MC202,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC204",  dsPIC33EP256MC204,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC206",  dsPIC33EP256MC206,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC502",  dsPIC33EP256MC502,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC504",  dsPIC33EP256MC504,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP256MC506",  dsPIC33EP256MC506,  88064, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512GP202",    PIC24EP512GP202,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512GP204",    PIC24EP512GP204,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512GP206",    PIC24EP512GP206,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512GP502",  dsPIC33EP512GP502,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512GP504",  dsPIC33EP512GP504,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512GP506",  dsPIC33EP512GP506,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512MC202",    PIC24EP512MC202,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512MC204",    PIC24EP512MC204,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"PIC24EP512MC206",    PIC24EP512MC206,    175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC202",  dsPIC33EP512MC202,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC204",  dsPIC33EP512MC204,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC206",  dsPIC33EP512MC206,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC502",  dsPIC33EP512MC502,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC504",  dsPIC33EP512MC504,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},
{"dsPIC33EP512MC506",  dsPIC33EP512MC506,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,1024,2,   0,0,0},

{"dsPIC33EP128GM304",  dsPIC33EP128GM304,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP128GM604",  dsPIC33EP128GM604,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP128GM306",  dsPIC33EP128GM306,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP128GM706",  dsPIC33EP128GM706,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP128GM310",  dsPIC33EP128GM310,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP128GM710",  dsPIC33EP128GM710,  44032, 0,     DS70663D,  0x0157EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM304",  dsPIC33EP256GM304,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM604",  dsPIC33EP256GM604,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM306",  dsPIC33EP256GM306,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM706",  dsPIC33EP256GM706,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM310",  dsPIC33EP256GM310,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP256GM710",  dsPIC33EP256GM710,  87552, 0,     DS70663D,  0x02AFEC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM304",  dsPIC33EP512GM304,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM604",  dsPIC33EP512GM604,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM306",  dsPIC33EP512GM306,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM706",  dsPIC33EP512GM706,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM310",  dsPIC33EP512GM310,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},
{"dsPIC33EP512GM710",  dsPIC33EP512GM710,  175104,0,     DS70663D,  0x0557EC,10, 0x800FF8, 4, 0, 0,        2,0,512,2,    0,0,0},

{"dsPIC30F1010",       dsPIC30F1010,       2048,  0,     DS70284B,  0xF80000, 8, 0x8005C0,32, 0, 0,        32,0,32,0,    0,0,0},
{"dsPIC30F2020",       dsPIC30F2020,       4096,  0,     DS70284B,  0xF80000, 8, 0x8005C0,32, 0, 0,        32,0,32,0,    0,0,0},
{"dsPIC30F2023",       dsPIC30F2023,       4096,  0,     DS70284B,  0xF80000, 8, 0x8005C0,32, 0, 0,        32,0,32,0,    0,0,0},

{"PIC24FJ16MC101",     PIC24FJ16MC101,     5632,  0,     DS75012B,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"PIC24FJ16MC102",     PIC24FJ16MC102,     5632,  0,     DS75012B,  0x002BFC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"PIC24FJ32MC101",     PIC24FJ32MC101,     11264, 0,     DS75012B,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"PIC24FJ32MC102",     PIC24FJ32MC102,     11264, 0,     DS75012B,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},
{"PIC24FJ32MC104",     PIC24FJ32MC104,     11264, 0,     DS75012B,  0x0057FC, 2, 0, 0,        0, 0,        1,0,512,0,    0,0,0},

{"PIC24FJ32GA102",     PIC24FJ32GA102,     11264, 0,     DS39934B,  0x0057F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA102",     PIC24FJ64GA102,     22016, 0,     DS39934B,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ32GA104",     PIC24FJ32GA104,     11264, 0,     DS39934B,  0x0057F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA104",     PIC24FJ64GA104,     22016, 0,     DS39934B,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ32GB002",     PIC24FJ32GB002,     11264, 0,     DS39934B,  0x0057F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB002",     PIC24FJ64GB002,     22016, 0,     DS39934B,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ32GB004",     PIC24FJ32GB004,     11264, 0,     DS39934B,  0x0057F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB004",     PIC24FJ64GB004,     22016, 0,     DS39934B,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},

/* These devices have 3 config words at the end of program flash */
{"PIC24FJ64GA106",     PIC24FJ64GA106,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA108",     PIC24FJ64GA108,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA110",     PIC24FJ64GA110,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB106",     PIC24FJ64GB106,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB108",     PIC24FJ64GB108,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB110",     PIC24FJ64GB110,     22016, 0,  DS30009907C,  0x00ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA106",    PIC24FJ128GA106,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA108",    PIC24FJ128GA108,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA110",    PIC24FJ128GA110,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB106",    PIC24FJ128GB106,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB108",    PIC24FJ128GB108,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB110",    PIC24FJ128GB110,    44032, 0,  DS30009907C,  0x0157FA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GA106",    PIC24FJ192GA106,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GA108",    PIC24FJ192GA108,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GA110",    PIC24FJ192GA110,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GB106",    PIC24FJ192GB106,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GB108",    PIC24FJ192GB108,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ192GB110",    PIC24FJ192GB110,    67072, 0,  DS30009907C,  0x020BFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GA106",    PIC24FJ256GA106,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GA108",    PIC24FJ256GA108,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GA110",    PIC24FJ256GA110,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GB106",    PIC24FJ256GB106,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GB108",    PIC24FJ256GB108,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GB110",    PIC24FJ256GB110,    87552, 0,  DS30009907C,  0x02ABFA, 3, 0,0,         0x8007F0, 8, 64,0,512,0,   0,0,0},

/* These devices have 4 config words at the end of program flash */
{"PIC24FJ64GA204",     PIC24FJ64GA204,     22016, 0,  DS30000510E,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA202",     PIC24FJ64GA202,     22016, 0,  DS30000510E,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB204",     PIC24FJ64GB204,     22016, 0,  DS30000510E,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GB202",     PIC24FJ64GB202,     22016, 0,  DS30000510E,  0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA204",    PIC24FJ128GA204,    44032, 0,  DS30000510E,  0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA202",    PIC24FJ128GA202,    44032, 0,  DS30000510E,  0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB204",    PIC24FJ128GB204,    44032, 0,  DS30000510E,  0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB202",    PIC24FJ128GB202,    44032, 0,  DS30000510E,  0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},

/* These devices have 4 config words at the end of program flash */
{"PIC24FJ64GA306",     PIC24FJ64GA306,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA308",     PIC24FJ64GA308,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GA310",     PIC24FJ64GA310,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GC006",     PIC24FJ64GC006,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GC008",     PIC24FJ64GC008,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ64GC010",     PIC24FJ64GC010,     22016, 0,    DS39970E,   0x00ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128DA106",    PIC24FJ128DA106,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128DA110",    PIC24FJ128DA110,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128DA206",    PIC24FJ128DA206,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128DA210",    PIC24FJ128DA210,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA306",    PIC24FJ128GA306,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA308",    PIC24FJ128GA308,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GA310",    PIC24FJ128GA310,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB206",    PIC24FJ128GB206,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GB210",    PIC24FJ128GB210,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GC006",    PIC24FJ128GC006,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GC008",    PIC24FJ128GC008,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ128GC010",    PIC24FJ128GC010,    44032, 0,    DS39970E,   0x0157F8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256DA106",    PIC24FJ256DA106,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256DA110",    PIC24FJ256DA110,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256DA206",    PIC24FJ256DA206,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256DA210",    PIC24FJ256DA210,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GB206",    PIC24FJ256GB206,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},
{"PIC24FJ256GB210",    PIC24FJ256GB210,    87552, 0,    DS39970E,   0x02ABF8, 4, 0,0,         0x800880, 8, 64,0,512,0,   0,0,0},

/* These devices use 10 of 25 config words in the last page of program flash and support dual partition mode */
{"PIC24FJ64GA406",     PIC24FJ64GA406,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ64GA410",     PIC24FJ64GA410,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ64GA412",     PIC24FJ64GA412,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ64GB406",     PIC24FJ64GB406,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ64GB410",     PIC24FJ64GB410,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ64GB412",     PIC24FJ64GB412,     22528, 0,   DS30010073A, 0x00AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x005780,0},
{"PIC24FJ128GA406",    PIC24FJ128GA406,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ128GA410",    PIC24FJ128GA410,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ128GA412",    PIC24FJ128GA412,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ128GB406",    PIC24FJ128GB406,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ128GB410",    PIC24FJ128GB410,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ128GB412",    PIC24FJ128GB412,    44032, 0,   DS30010073A, 0x015780, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x00AB80,0},
{"PIC24FJ256GA406",    PIC24FJ256GA406,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},
{"PIC24FJ256GA410",    PIC24FJ256GA410,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},
{"PIC24FJ256GA412",    PIC24FJ256GA412,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},
{"PIC24FJ256GB406",    PIC24FJ256GB406,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},
{"PIC24FJ256GB410",    PIC24FJ256GB410,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},
{"PIC24FJ256GB412",    PIC24FJ256GB412,    88064, 0,   DS30010073A, 0x02AF80, 25, 0,0,        0, 0,        64,0,512,64,PIC24_FBOOT,0x015780,0},

/* These devices use 9 of 23 config words in the last page of program flash and support dual partition mode */
{"PIC24FJ128GA606",    PIC24FJ128GA606,    45056,  0,  DS30010057F, 0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x00AF00,0},
{"PIC24FJ128GA610",    PIC24FJ128GA610,    45056,  0,  DS30010057F, 0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x00AF00,0},
{"PIC24FJ128GB606",    PIC24FJ128GB606,    45056,  0,  DS30010057F, 0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x00AF00,0},
{"PIC24FJ128GB610",    PIC24FJ128GB610,    45056,  0,  DS30010057F, 0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x00AF00,0},
{"PIC24FJ256GA606",    PIC24FJ256GA606,    88064,  0,  DS30010057F, 0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x015700,0},
{"PIC24FJ256GA610",    PIC24FJ256GA610,    88064,  0,  DS30010057F, 0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x015700,0},
{"PIC24FJ256GB606",    PIC24FJ256GB606,    88064,  0,  DS30010057F, 0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x015700,0},
{"PIC24FJ256GB610",    PIC24FJ256GB610,    88064,  0,  DS30010057F, 0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x015700,0},
{"PIC24FJ512GA606",    PIC24FJ512GA606,    176128, 0,  DS30010057F, 0x055F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x02AF00,0},
{"PIC24FJ512GA610",    PIC24FJ512GA610,    176128, 0,  DS30010057F, 0x055F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x02AF00,0},
{"PIC24FJ512GB606",    PIC24FJ512GB606,    176128, 0,  DS30010057F, 0x055F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x02AF00,0},
{"PIC24FJ512GB610",    PIC24FJ512GB610,    176128, 0,  DS30010057F, 0x055F00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x02AF00,0},
{"PIC24FJ1024GA606",   PIC24FJ1024GA606,   352256, 0,  DS30010057F, 0x0ABF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x055F00,0},
{"PIC24FJ1024GA610",   PIC24FJ1024GA610,   352256, 0,  DS30010057F, 0x0ABF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x055F00,0},
{"PIC24FJ1024GB606",   PIC24FJ1024GB606,   352256, 0,  DS30010057F, 0x0ABF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x055F00,0},
{"PIC24FJ1024GB610",   PIC24FJ1024GB610,   352256, 0,  DS30010057F, 0x0ABF00, 23, 0,0,        0, 0,       128,0,1024,128,PIC24_FBOOT,0x055F00,0},

/* These devices use 15 of 35 config words in the last page of program flash */
{"dsPIC33EV32GM002",   dsPIC33EV32GM002,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV32GM004",   dsPIC33EV32GM004,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV32GM006",   dsPIC33EV32GM006,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV32GM102",   dsPIC33EV32GM102,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV32GM104",   dsPIC33EV32GM104,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV32GM106",   dsPIC33EV32GM106,   11264, 0,  DS70005137F,  0x005780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM002",   dsPIC33EV64GM002,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM004",   dsPIC33EV64GM004,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM002",   dsPIC33EV64GM002,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM104",   dsPIC33EV64GM104,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM106",   dsPIC33EV64GM106,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV64GM106",   dsPIC33EV64GM106,   22016, 0,  DS70005137F,  0x00AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM002",  dsPIC33EV128GM002,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM004",  dsPIC33EV128GM004,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM006",  dsPIC33EV128GM006,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM102",  dsPIC33EV128GM102,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM104",  dsPIC33EV128GM104,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV128GM106",  dsPIC33EV128GM106,  44032, 0,  DS70005137F,  0x015780, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM002",  dsPIC33EV256GM002,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM004",  dsPIC33EV256GM004,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM006",  dsPIC33EV256GM006,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM102",  dsPIC33EV256GM102,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM104",  dsPIC33EV256GM104,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},
{"dsPIC33EV256GM106",  dsPIC33EV256GM106,  87552, 0,  DS70005137F,  0x02AB80, 35, 0,0,        0, 0,        2,0,512,2,   0,0,64},

/* These devices use 10 of 25 config words in the last page of program flash and support dual partition mode (64KB only) */
{"dsPIC33EP16GS502",   dsPIC33EP16GS502,   5632,  0,  DS70005160C,  0x002B80, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP16GS504",   dsPIC33EP16GS504,   5632,  0,  DS70005160C,  0x002B80, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP16GS505",   dsPIC33EP16GS505,   5632,  0,  DS70005160C,  0x002B80, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP16GS506",   dsPIC33EP16GS506,   5632,  0,  DS70005160C,  0x002B80, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP32GS502",   dsPIC33EP32GS502,   11264, 0,  DS70005160C,  0x005780, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP32GS504",   dsPIC33EP32GS504,   11264, 0,  DS70005160C,  0x005780, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP32GS505",   dsPIC33EP32GS505,   11264, 0,  DS70005160C,  0x005780, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP32GS506",   dsPIC33EP32GS506,   11264, 0,  DS70005160C,  0x005780, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP64GS502",   dsPIC33EP64GS502,   22528, 0,  DS70005160C,  0x00AF80, 25, 0,0,        0, 0,        2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS504",   dsPIC33EP64GS504,   22528, 0,  DS70005160C,  0x00AF80, 25, 0,0,        0, 0,        2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS505",   dsPIC33EP64GS505,   22528, 0,  DS70005160C,  0x00AF80, 25, 0,0,        0, 0,        2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS506",   dsPIC33EP64GS506,   22528, 0,  DS70005160C,  0x00AF80, 25, 0,0,        0, 0,        2,0,512,2,dsPIC_FBOOT,0x5780,64},

/* These devices use 10 of 25 config words in the last page of program flash */
{"dsPIC33EP16GS202",   dsPIC33EP16GS202,   5632,  0,  DS70005192B,  0x002B80, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},
{"dsPIC33EP32GS202",   dsPIC33EP32GS202,   11264, 0,  DS70005192B,  0x005780, 25, 0,0,        0, 0,        2,0,512,2,  0,0,64},

/* These devices use 9 of 23 config words in the last page of program flash */
{"PIC24FJ64GA702",     PIC24FJ64GA702,     22528, 0,  DS30010102A,  0x00AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ64GA704",     PIC24FJ64GA704,     22528, 0,  DS30010102A,  0x00AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ64GA705",     PIC24FJ64GA705,     22528, 0,  DS30010102A,  0x00AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ128GA702",    PIC24FJ128GA702,    45056, 0,  DS30010102A,  0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ128GA704",    PIC24FJ128GA704,    45056, 0,  DS30010102A,  0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ128GA705",    PIC24FJ128GA705,    45056, 0,  DS30010102A,  0x015F00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ256GA702",    PIC24FJ256GA702,    88064, 0,  DS30010102A,  0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ256GA704",    PIC24FJ256GA704,    88064, 0,  DS30010102A,  0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},
{"PIC24FJ256GA705",    PIC24FJ256GA705,    88064, 0,  DS30010102A,  0x02AF00, 23, 0,0,        0, 0,       128,0,1024,128,0,0,0},

{"dsPIC33EP64GS708",   dsPIC33EP64GS708,   22528, 0,  DS70005256A,  0x00AF80, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS804",   dsPIC33EP64GS804,   22528, 0,  DS70005256A,  0x00AF80, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS805",   dsPIC33EP64GS805,   22528, 0,  DS70005256A,  0x00AF80, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS806",   dsPIC33EP64GS806,   22528, 0,  DS70005256A,  0x00AF80, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP64GS808",   dsPIC33EP64GS808,   22528, 0,  DS70005256A,  0x00AF80, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0x5780,64},
{"dsPIC33EP128GS702",  dsPIC33EP128GS702,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS704",  dsPIC33EP128GS704,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS705",  dsPIC33EP128GS705,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS706",  dsPIC33EP128GS706,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS708",  dsPIC33EP128GS708,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS804",  dsPIC33EP128GS804,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS805",  dsPIC33EP128GS805,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS806",  dsPIC33EP128GS806,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},
{"dsPIC33EP128GS808",  dsPIC33EP128GS808,  44032, 0,  DS70005256A,  0x015780, 25, 0,0,        0, 0,       2,0,512,2,dsPIC_FBOOT,0xAB80,64},

{"(null)",             0,                  0,     0,  0,            0,        0,  0,0,        0, 0,       0,0,0,0,   0,0,0},
/*Device name          Device id           Flash  EEPROM Data-sheet Config    #   Fuid #      Calib #     PanelSizes BOOT/Config/MEM*/
};
#define PIC24_SIZE ((sizeof(pic24_map) / sizeof(struct pic24_dsmap)) - 1)

/* Default device (null) */
uint32_t pic24_index = PIC24_SIZE;

void
pic24_selector(void)
{
	uint32_t i;
	char *dnames[PIC24_SIZE];

	for (i = 0; i < PIC24_SIZE; ++i)
		dnames[i] = pic24_map[i].devicename;

	qsort(dnames, PIC24_SIZE, sizeof(char *), pic_cmp);

	for (i = 0; i < PIC24_SIZE; ++i) {
		if ((i % PIC_NCOLS) == (PIC_NCOLS - 1))
			printf("%s\n", dnames[i]);
		else
			printf("%-19s", dnames[i]);
	}
	if (i % PIC_NCOLS)
		printf("\n");

	printf("Total: %u\n", (uint32_t)PIC24_SIZE);
}

#ifdef LOADER
void
pic24_bootloader(void)
{
	uint32_t dev, flerase = 0, flwrite = 0, eeerase = 0, eewrite = 0;
	char *devname;

	for (dev = 0; pic24_map[dev].deviceid; ++dev) {
		if (pic24_map[dev].devicename[0] == 'P')
			devname = &pic24_map[dev].devicename[3];
		else
			devname = &pic24_map[dev].devicename[5];
		printf(".ifdef __%s\n", devname);
		printf("    .include    \"p%s.inc\"\n", devname);
		if (devname[1] == '0') {	/* 30F */
			flerase = 0x4041;
			flwrite = 0x4001; 	/* 1 ROW */
			eeerase = 0x4044;
			eewrite = 0x4004; 	/* 1 ROW */
		}
		else if (devname[2] == 'E') {	/* 33EP, 33EV or 24EP */
			flerase = 0x4003;
			if (pic24_map[dev].mempanelsize)
				flwrite = 0x4202;/* 1 ROW PACKED    */
			else
				flwrite = 0x4001;/* 2 WORDS LATCHED */
		}
		else if (devname[2] == 'F') {	/* 33F or 24F */
			if (devname[3] != 'J') {  /* 24F K */
				flerase = 0x4058;
				flwrite = 0x4004; /* 1 ROW */
				eeerase = 0x4058;
				eewrite = 0x4004; /* 1 ROW */
			} else if (pic24_map[dev].codepanelsize == 1) {
				flerase = 0x4042;
				flwrite = 0x4003; /* 1 WORD */
			} else if (pic24_map[dev].latch) {
				flerase = 0x4003;
				flwrite = 0x4002; /* 1 ROW LATCHED */
			} else {
				flerase = 0x4042;
				flwrite = 0x4001; /* 1 ROW */
			}
		}
		else if (devname[2] == 'H') {	/* 24H */
			flerase = 0x4042;
			flwrite = 0x4001; 	/* 1 ROW */
		}
		printf("    .equiv      MAX_FLASH,  0x%X\n", pic24_map[dev].flash << 1);
		if (pic24_map[dev].eeprom)
		printf("    .equiv      MAX_EE,     %d ; X 64\n", pic24_map[dev].eeprom / 64);
		if (pic24_map[dev].mempanelsize) {
			printf("    .equiv      ROWSIZE,    %d\n", pic24_map[dev].mempanelsize);
			printf("    .equiv      ROWREPEAT,  1\n");	/* NO REPEAT */
			printf("    .equiv      SRCPACKED,  1\n");	/* PACKED  */
		} else {
			printf("    .equiv      ROWSIZE,    %d\n", pic24_map[dev].codepanelsize);
			if (pic24_map[dev].codepanelsize == 1) {
			printf("    .equiv      ROWREPEAT,  64\n");	/*  64 WORDS TOTAL (RAM LIMIT) */
			} else if (pic24_map[dev].codepanelsize == 2) {
			printf("    .equiv      ROWREPEAT,  64\n");	/* 128 WORDS TOTAL (MAX LIMIT) */
			} else {
			printf("    .equiv      ROWREPEAT,  1\n");	/* NO REPEAT */
			}
			printf("    .equiv      SRCPACKED,  0\n");	/* NOT PACKED */
		}
		if (pic24_map[dev].latch)
		printf("    .equiv      LATCHSIZE,  %d\n", pic24_map[dev].latch);
		printf("    .equiv      ERASESIZE,  %d\n", pic24_map[dev].erasepanelsize);
		printf("    .equiv      ERASE_FLASH,0x%04X\n", flerase);
		printf("    .equiv      WRITE_FLASH,0x%04X\n", flwrite);
		if (pic24_map[dev].eeprom) {
		printf("    .equiv      ERASE_EE,   0x%04X\n", eeerase);
		printf("    .equiv      WRITE_EE,   0x%04X\n", eewrite);
		}
		printf(".endif\n");
	}
}
#endif

/*****************************************************************************
 *
 * Hardware algorithm map: Device family
 *
 *****************************************************************************/
#if 0
struct pic24_dstab pic24_tab[] =
{
	/* EOF */
	{0, "(null)"},
};
#endif

/*****************************************************************************
 *
 * Program/Verify mode
 *
 *****************************************************************************/

/*
 * ENTER HVP/LVP PROGRAM/VERIFY MODE
 *
 * ENTRY - VDD FIRST
 *
 * DS39768D-page 15 PIC24FJ64GA002
 * DS70102H-page 68 dsPIC30F4013
 *
 * NOTES: EXIT THE RESET VECTOR (SOME ASSUMPTIONS)
 *
 *  ON RESET THE DEVICE EFFECTIVELY PERFORMS A GOTO 0
 *  AND WE MAY ASSUME THAT WHAT IS CALLED THE `FORCED NOP'
 *  IN THE DOCUMENTATION IS THE EXECUTION OF THE GOTO 0
 *  INSTRUCTION.
 *
 *  ON MOST dsPIC/PIC24 DEVICES, GOTO REQUIRES TWO CYCLES BUT
 *  ON dsPIC33E/PIC24E IT DEMANDS FOUR, THIS MEANS ON RESET WE
 *  NEED A DIFFERENT NUMBER OF GOTO NOPS.
 *
 *  ON LVP DEVICES THE INTIAL `FORCED NOP' IS 4+1 CLOCK CYCLES
 *  RATHER THAN 4+24 LONG, HOWEVER, THIS ISN'T DOCUMENTED AS
 *  SUCH IN THE MICROCHIP DATA-SHEET AND THIS IS JUST AN
 *  OBSERVATION.
 *
 *  AFTER RESET WE PERFORM A SINGLE `FORCED NOP' (WHICH MAY
 *  BE SHORT) AND THAT IS THE GOTO 0 INSTRUCTION, NEXT WE
 *  PERFORM EITHER ONE NOP TO COMPLETE THE INSTRUCTION FOR
 *  dsPIC/PIC24 ELSE THREE NOPS FOR dsPIC33E/PIC24E.
 *
 *  ONCE THE FORCED NOPS ARE COMPLETED WE MUST GOTO
 *  CODE ORIGIN WHICH IS NORMALLY 200 BUT WAS 100 ON
 *  THE FIRST dsPIC/PIC24 DEVICES. FAILURE TO JUMP NORMALLY
 *  RESULTS IN A CPU CRASH.
 */
void
pic24_program_verify(void)
{
	/* RESET & ACQUIRE GPIO */
	io_set_vpp(LOW);
	/* DS39768D    PIC24FJ64GA002    P6(100ns) */
	/* DS70102H    dsPIC30F4013      P6(100ns) */
	/* DS70663D    dsPIC33EP128GP502 P6(100ns) */
	/* DS70005137F dsPIC33EV256GM002 P6(100ns) */
	/* DS30000510E PIC24FJXXGA2/GB2  P6(100ns) */
	io_usleep(1000);

	/* PGD + PGC + PGM(N/A) LOW */
	io_set_pgd(LOW);
	io_set_pgc(LOW);
	io_set_pgm(LOW);
	io_usleep(1000);

	/* LVP(KEY) */
	if (p.key == LVPKEY) {
		/* INPUT DATA ON CLOCK FALLING EDGE, LSB FIRST */
		/* DS39768D-page 14 PIC24FJ64GA002 */
		io_configure(TRUE, FALSE);

		/* VPP HIGH */
		io_set_vpp(HIGH);
		/* DS39768D    PIC24FJ64GA002    UNKNOWN    */
		/* DS70663D    dsPIC33EP128GP502 P21(500us) */
		/* DS70005137F dsPIC33EV256GM002 P21(500us) */
		/* DS30000510E PIC24FJXXGA2/GB2  UNKNOWN    */
		io_usleep(500);

		/* VPP LOW */
		io_set_vpp(LOW);
		/* DS39768D    PIC24FJ64GA002    P18(40ns) */
		/* DS70663D    dsPIC33EP128GP502 P18(1ms)  */
		/* DS70005137F dsPIC33EV256GM002 P18(1ms)  */
		/* DS30000510E PIC24FJXXGA2/GB2  P18(10ms) */
		io_usleep(10000);

		/* PROGRAM/VERIFY ENTRY CODE */
		io_program_out(QHCMKEY, 32);
		/* DS39768D    PIC24FJ64GA002    P19(1ms)  */
		/* DS70663D    dsPIC33EP128GP502 P19(25ns) */
		/* DS70005137F dsPIC33EV256GM002 P19(25ns) */
		/* DS30000510E PIC24FJXXGA2/GB2  P19(1ms)  */
		io_usleep(1000);

		/* VPP HIGH */
		io_set_vpp(HIGH);
		/* DS39768D    PIC24FJ64GA002    P7(25ms)      */
		/* DS70663D    dsPIC33EP128GP502 P7(50ms)  +   */
		/* DS70663D    dsPIC33EP128GP502 P1(200ns) x 5 */
		/* DS30000510E PIC24FJXXGA2/GB2  P7(25ms)      */
		io_usleep(50001);

		/*
		 * STEP 1: EXIT THE RESET VECTOR (GOTO 0)
		 * 	NOP (5)
		 *	NOP
		 *	NOP, NOP (dsPIC33E/PIC24E)
		 *	GOTO 200 (CODE ORIGIN)
		 *	NOP
		 *	NOP, NOP (dsPIC33E/PIC24E)
		 */

		/* DS39768D-page 13 PIC24FJ64GA002 */
		/* NOP (5)              */
		/* NOP (28) [,NOP, NOP] */
		uint8_t nops = pic24_conf.gotonop;

		io_program_out(0, 5);
		while (nops--)
			io_program_out(0, 28);	/* NOP */

		/* GOTO 200, NOP [,NOP, NOP] */
		pic24_goto200();
	}
	/* HVP(KEY) */
	else if (p.key == HVPKEY) {
		/* INPUT DATA ON CLOCK FALLING EDGE, LSB FIRST */
		/* DS30625D-page 9 PIC24F16KL402 */
		io_configure(TRUE, FALSE);

		/* VPP HIGH */
		io_set_vpp(HIGH);
		/* DS30625D PIC24F16KL402  P18(40ns) */
		/* DS39919B PIC24FV32KA302 P18(1ms)  */
		io_usleep(1000);

		/* PROGRAM/VERIFY ENTRY CODE */
		io_program_out(QHCMKEY, 32);
		/* DS30625D PIC24F16KL402  P7(25ms) */
		/* DS39919B PIC24FV32KA302 P7(25ms) */
		io_usleep(25000); /* 25ms */

		/*
		 * STEP 1: EXIT THE RESET VECTOR (GOTO 0)
		 *	NOP (5)
		 *	NOP
		 *	GOTO 200 (CODE ORIGIN)
		 *	NOP
		 */

		/* DS39625D-page 10 PIC24F16KL402 */
		/* NOP (5)  */
		/* NOP (28) */
		io_program_out(0, 5 + 28);

		/* GOTO 200, NOP */
		pic24_goto200();
	}
	/* HVP(STDP) */
	else {
		/* INPUT DATA ON CLOCK RISING EDGE, LSB FIRST */
		io_configure(FALSE, FALSE);

		/* VPP HIGH */
		io_set_vpp(HIGH);
		/* DS70102H dsPIC30F4013 UNNAMED(25ms) */
		io_usleep(25000); /* 25ms */

		/*
		 * STEP 1: EXIT THE RESET VECTOR (GOTO 0)
		 *	NOP
		 *	NOP
		 *	GOTO 100 (CODE ORIGIN)
		 *	NOP
		 */

		/* DS70102H-page 69 dsPIC30F4013 */
		/* NOP (28) */
		/* NOP (28) */
		io_program_out(0, 28 + 28);

		/* GOTO 200, NOP */
		pic24_goto200();
	}
}

/*
 * EXIT HVP/LVP PROGRAM/VERIFY MODE
 *
 * EXIT - VDD LAST
 */
void
pic24_standby(void)
{
	/* PGD + PGC + VPP + PGM(N/A) LOW */
	io_set_pgd(LOW);
	io_set_pgc(LOW);
	io_set_vpp(LOW);
	io_set_pgm(LOW);
}

/*****************************************************************************
 *
 * Partition status functions
 *
 *****************************************************************************/

/*
 * GET PARTITION SUPPORT STATUS (FROM DEVICE TABLE)
 */
static inline int
pic24_supports_partition(void)
{
	if (pic24_map[pic24_index].fbootaddr == 0 || pic24_map[pic24_index].configlower == 0)
		return 0;

	return 1;
}

/*
 * GET DESIRED PARTITION STATUS (TO CHIP)
 */
static inline int
pic24_wants_partition(void)
{
	if (pic24_map[pic24_index].fbootaddr == 0 || pic24_map[pic24_index].configlower == 0)
		return 0;

	return p.partition;
}

/*
 * GET CURRENT PARTITION STATUS (FROM CHIP)
 */
static inline int
pic24_has_partition(void)
{
	if (pic24_map[pic24_index].fbootaddr == 0 || pic24_map[pic24_index].configlower == 0)
		return 0;

	return (pic24_conf.fboot & PIC24_FBOOT_MASK) != PIC24_SINGLE;
}

/*****************************************************************************
 *
 * Hardware functions
 *
 *****************************************************************************/

/*
 * 0000
 *
 * SIX (SHIFT IN 24-BIT INSTRUCTION AND EXECUTE)
 */
static inline void
pic24_six(uint32_t data, uint8_t nops)
{
	io_program_out(0x0, 4);		/* SIX */
	io_program_out(data, 24);
	while (nops--)
		io_program_out(0, 28);	/* NOP */
}

/*
 * 0001
 *
 * REGOUT (SHIFT OUT THE VISI REGISTER)
 */
static inline uint16_t
pic24_regout(void)
{
	uint16_t word;

	io_program_out(0x1, 4);		/* REGOUT */
	word = io_program_in(24) >> 8;	/* VISI   */
	io_program_out(0, 28);		/* NOP    */

	return word;
}

/*
 * 0010 .. 1111
 *
 * RESERVED
 */

/*****************************************************************************
 *
 * Compound functions
 *
 *****************************************************************************/

/*
 * GOTO 200
 *
 *  GO TO CODE ORIGIN
 */
void
pic24_goto200(void)
{
	pic24_six(0x040200, pic24_conf.gotonop); /* GOTO 200 */
}

/*
 * READ REGISTER
 */
uint16_t
pic24_readreg(uint16_t regin)
{
	uint32_t reg  = 0x800000 | (regin << 3);
	uint32_t visi = 0x880000 | (pic24_conf.visi << 3);
	uint16_t regout;

	pic24_six(reg,  1);	/* MOV #reg, W0 */
	pic24_six(visi, 1);	/* MOV W0, VISI */

	regout = pic24_regout();

	return regout;
}

/*
 * INITIALISE TBLPAG, THE READ POINTER (W6) AND THE WRITE POINTER (W7) FOR TBLRD INSTRUCTION
 *
 *  AFFECTS
 *   TBLPAG = ADDRESS HIGH
 *   W6 = ADDRESS LOW
 *   W7 = VISI
 */
void
pic24_set_read_pointer(uint32_t addr)
{
	uint32_t high   = 0x200000 | ((addr & 0x00FF0000) >> 12);
	uint32_t tblpag = 0x880000 | (pic24_conf.tblpag   << 3);
	uint32_t low    = 0x200006 | ((addr & 0x0000FFFF) << 4);
	uint32_t visi   = 0x200007 | (pic24_conf.visi     << 4);
#ifdef DEBUG
	printf("%s: PCL=%04X\n", __func__, pic24_readreg(_PCL));
#endif
	pic24_six(high,   0);	/* MOV #<SourceAddress23:16>, W0 */
	pic24_six(tblpag, 0);	/* MOV W0, TPLPAG                */
	pic24_six(low,    0);	/* MOV #<SourceAddress15:0>, W6  */
	pic24_six(visi,   1);	/* MOV #VISI, W7                 */
}

/*
 * INITIALISE TBLPAG AND THE WRITE POINTER (W7) FOR TBLWT INSTRUCTION
 *
 *  AFFECTS
 *   TBLPAG = ADDRESS HIGH
 *   W7 = ADDRESS LOW
 */
void
pic24_set_write_pointer(uint32_t addr)
{
	uint32_t high   = 0x200000 | ((addr & 0x00FF0000) >> 12);
	uint32_t tblpag = 0x880000 | (pic24_conf.tblpag   << 3);
	uint32_t low    = 0x200007 | ((addr & 0x0000FFFF) << 4);
#ifdef DEBUG
	printf("%s: PCL=%04X\n", __func__, pic24_readreg(_PCL));
#endif
	pic24_six(high,   0);	/* MOV #<DestinationAddress23:16>, W0 */
	pic24_six(tblpag, 0);	/* MOV W0, TPLPAG                     */
	pic24_six(low,    1);	/* MOV #<DestinationAddress15:0>, W7  */
}

/*
 * Set the NVMADRU/NVMADR register pair to point to the correct address.
 *
 *  AFFECTS
 *   NVMADR = ADDRESS LOW
 *   NVMADRU = ADDRESS HIGH
 *   W7 = RESET
 */
void
pic24_set_write_pointer_nvmadr(uint32_t addr)
{
	uint32_t high = 0x200004 | ((addr & 0x00FF0000) >> 12);
	uint32_t low  = 0x200003 | ((addr & 0x0000FFFF) << 4);
#ifdef DEBUG
	printf("%s: PCL=%04X\n", __func__, pic24_readreg(_PCL));
#endif
	pic24_six(low, 1);	/* MOV #<DestinationAddress15:0>, W3  */
	pic24_six(high, 0);	/* MOV #<DestinationAddress23:16>, W4 */
	pic24_six(0x883B13, 0);	/* MOV W3, NVMADR                     */
	pic24_six(0x883B24, 0);	/* MOV W4, NVMADRU                    */
	pic24_six(0xEB0380, 1);	/* CLR W7                             */
}

/*
 * TABLE READ 16-BIT WITH POST INCREMENT
 *
 *  REQUIRES
 *   TBLPAG = ADDRESS HIGH
 *   W6 = ADDRESS LOW
 *   W7 = VISI
 *
 *   96 0B BA TBLRDL [W6], [W7]
 *   B6 0B BA TBLRDL [W6++], [W7]
 */
static inline uint16_t
pic24_table_read16_post_increment(void)
{
	uint16_t word;

	/* TABLE READ LOW (WORD) [TBLPAG]:[W6] => [W7] (VISI), INCREMENT W6            */
	pic24_six(0xBA0BB6, pic24_conf.tblnop);	/* TBLRDL [W6++], [W7]                 */
	word = pic24_regout();			/* Clock out contents of VISI register */

	return word;
}

/*
 * TABLE READ 24-BIT WITH POST INCREMENT
 *
 *  REQUIRES
 *   TBLPAG = ADDRESS HIGH
 *   W6 = ADDRESS LOW
 *   W7 = VISI
 *
 *   96 0B BA TBLRDL [W6], [W7]
 *   B6 0B BA TBLRDL [W6++], [W7]
 *
 *   96 8B BA TBLRDH [W6], [W7]
 *   B6 8B BA TBLRDH [W6++], [W7]
 */
uint32_t
pic24_table_read24_post_increment(void)
{
	uint32_t word;
	uint16_t low, high;

	/* TABLE READ LOW  WORD [TBLPAG]:[W6] => [W7] (VISI)                           */
	pic24_six(0xBA0B96, pic24_conf.tblnop);	/* TBLRDL [W6], [W7]                   */
	low = pic24_regout();			/* Clock out contents of VISI register */

	/* TABLE READ HIGH WORD [TBLPAG]:[W6++] => [W7] (VISI)                         */
	pic24_six(0xBA8BB6, pic24_conf.tblnop);	/* TBLRDH [W6++], [W7]                 */
	high = pic24_regout();			/* Clock out contents of VISI register */

	word = ((high << 16) | low) & PIC24_MASK;

	return word;
}

/*
 * TABLE READ 2 x 24-BIT WITH POST INCREMENT
 *
 *  REQUIRES
 *   TBLPAG = ADDRESS HIGH
 *   W6 = ADDRESS LOW
 *   W7 = VISI
 */
void
pic24_table_read48_post_increment(uint32_t *word0, uint32_t *word1)
{
	uint16_t msb;

	pic24_six(0xBA0B96, pic24_conf.tblnop);	/* TBLRDL [W6], [W7]                   */
	*word0 = pic24_regout();		/* Clock out contents of VISI register */

	pic24_six(0xBADBB6, pic24_conf.tblnop);	/* TBLRDH.B [W6++], [W7++]             */
	pic24_six(0xBAD3D6, pic24_conf.tblnop);	/* TBLRDH.B [++W6], [W7--]             */
	msb = pic24_regout();			/* Clock out contents of VISI register */

	pic24_six(0xBA0BB6, pic24_conf.tblnop);	/* TBLRDL [W6++], [W7]                 */
	*word1 = pic24_regout();		/* Clock out contents of VISI register */

	*word0 |= (msb & 0x00FF) << 16;
	*word1 |= (msb & 0xFF00) << 8;
}

/*
 * WAIT FOR PROGRAM/ERASE COMPLETION
 */
void
pic24_nvmcon_write_completion(uint16_t mask, uint16_t word)
{
	struct timeval tv1 = {0}, tv2 = {0}, tv3 = {0};
	uint16_t regout;

	gettimeofday(&tv1, NULL);
	io_program_out(0, 28); /* NOP */
	do {
		gettimeofday(&tv2, NULL);
		timersub(&tv2, &tv1, &tv3);

		regout = pic24_readreg(pic24_conf.nvmcon) & mask;
		pic24_goto200(); /* GOTO 200 */
	}
	while ((regout != word) && (tv3.tv_sec < PIC_TIMEOUT));

	if (regout != word) {
		printf("%s: fatal error: program/erase timed out.\n", __func__);
		pic24_standby();
		io_exit(EX_SOFTWARE); /* Panic */
	}
}

/*
 * INITIATE PROGRAM/ERASE THEN DELAY FOR COMPLETION
 *
 * DS70102H-page 70 dsPIC30F4013
 * DS70284C-page 30 dsPIC30F1010
 */
void
pic24_nvmcon_write_external(uint32_t tdly)
{
 	/* NVMCON = 0x760 */
	pic24_six(0x200558, 0);	/* MOV #0x55, W8      */
	pic24_six(0x200AA9, 0);	/* MOV #0xAA, W9      */
	pic24_six(0x883B38, 0);	/* MOV W8, NVMKEY     */
	pic24_six(0x883B39, 0);	/* MOV W9, NVMKEY     */
	pic24_six(0xA8E761, 4);	/* BSET NVMCON, #WR   */
	io_usleep(tdly);
	pic24_six(0xA9E761, 4);	/* BCLR NVMCON, #WR   */
	pic24_goto200();
}

/*
 * INITIATE PROGRAM/ERASE
 *
 * DS30010057F-page 19 PIC24FJXXGA6/GB6
 */
static inline void
pic24_nvmcon_write_initiate_GAB(void)
{
	pic24_six(0x200550, 0);	/* MOV #0x55, W0      */
	pic24_six(0x883B30, 0);	/* MOV W0, NVMKEY     */
	pic24_six(0x200AA0, 0);	/* MOV #0xAA, W0      */
	pic24_six(0x883B30, 0);	/* MOV W0, NVMKEY     */
}

/*
 * INITIATE PROGRAM/ERASE THEN WAIT FOR COMPLETION
 *
 * DS39768D-page 17 PIC24FJ64GA002
 */
static inline void
pic24_nvmcon_write_internal(void)
{
 	/* NVMCON = 0x760 */
	pic24_six(0xA8E761, 2); /* BSET NVMCON, #WR */

	pic24_nvmcon_write_completion(0x8000, 0x0000);
}

/*
 * RESTORE CALIBRATION WORDS AFTER EXEC ERASED
 */
void
pic24_restore_calibration(void)
{
	if (pic24_map[pic24_index].calibaddr == 0)
		return; /* Not supported */
	
	pic24_write_program_init();

	pic_write_panel(PIC_PANEL_BEGIN, PIC_REGIONEXEC, pic24_map[pic24_index].codepanelsize);
	uint32_t address = pic24_map[pic24_index].calibaddr;
	for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i) {
		pic_write_panel(PIC_PANEL_UPDATE, address >> 1, pic24_conf.calib[i]);
		address += 2;
	}
	pic_write_panel(PIC_PANEL_END, PIC_VOID, PIC_VOID);
}

/*
 * BLANK dsPIC30
 *
 * DS70102H-page 69 dsPIC30F4013
 * DS70284C-page 28 dsPIC30F1010 EXTERNALLY TIMED(200ms)
 */
void
pic24_dsPIC30F_erase(uint32_t tdly)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();

	/* ERASE CONFIG */
	if (pic24_map[pic24_index].deviceid == dsPIC30F5011 ||
		pic24_map[pic24_index].deviceid == dsPIC30F5013) {
		pic24_conf.config[PIC24_PART_UPPER][3] = PIC24_CONFIG_PENDING; /* FBS = 0x0000 */
		pic24_conf.config[PIC24_PART_UPPER][4] = PIC24_CONFIG_PENDING; /* FSS = 0x0000 */
	}
	pic24_write_config_words();

	/* ERASE UNIT */
	pic24_write_program(pic24_map[pic24_index].fuidaddr, pic24_conf.fuid,
		pic24_map[pic24_index].nfuid);

	/* 407F ERASE CODE, DATA, EXEC & PROTECTION */
	pic24_six(0x2407FA, 0);	/* MOV #0x407F, W10 */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */

	/* EXTERNALLY TIMED */
	pic24_nvmcon_write_external(tdly);

	pic24_standby();
}

/*
 * BLANK PIC24FJ
 *
 * DS39768D-page 17    PIC24FJ64GA002
 * DS30000510E-page 14 PIC24FJXXGA2/GB2
 */
void
pic24_PIC24FJ_erase(uint32_t tblpag)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();

	/* 404F ERASE CHIP */
	pic24_six(0x2404FA, 0);	/* MOV #0x404F, W10   */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON    */
	pic24_six(0x200000, 0);	/* MOV #<PAGEVAL>, W0 */
	pic24_six(tblpag, 0);	/* MOV W0, TBLPAG     */
	pic24_six(0x200000, 0);	/* MOV #0x0000, W0    */
	pic24_six(0xBB0800, 2);	/* TBLWTL W0, [W0]    */

	pic24_nvmcon_write_internal();

	pic24_standby();
}

/*
 * BLANK dsPIC33FJ/PIC24FJXXMC
 *
 * DS70658C-page 12 dsPIC33FJ06GS101A (P11=240ms, P10=400ns)
 * DS75012B-page 12 PIC24FJ16MC101    (P11=240ms, P10=400ns)
 */
void
pic24_dsPIC33FJ_erase(void)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();

	/* 404F ERASE CHIP */
	pic24_six(0x2404FA, 0);	/* MOV #0x404F, W10     */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON      */
	pic24_six(0x200000, 0);	/* MOV #<PAGEVAL>, W0   */
	pic24_six(0x880190, 0);	/* MOV W0, TBLPAG       */
	pic24_six(0xA8E761, 4);	/* BSET NVMCON, #WR     */
	io_usleep(240001);	/* P11=240ms, P10=400ns */

	pic24_nvmcon_write_completion(0x8000, 0x0000);

	pic24_standby();
}

/*
 * BLANK PIC24FJXXGA4XX/PIC24FJXXGA6XX/PIC24FJXXGB6XX/PIC24FJXXGA7XX
 *
 * DS30010057F-page 19 PIC24FJ1024GA610/GB610 FAMILY
 */
void
pic24_PIC24FJ_GAB_erase(void)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();

	/* 400E ERASE CHIP */
	pic24_six(0x2400E0, 0);	/* MOV #0x400E, W0 */
	pic24_six(0x883B00, 0);	/* MOV W0, NVMCON  */
	pic24_nvmcon_write_initiate_GAB();
	pic24_six(0xA8E761, 4);	/* BSET NVMCON, #WR   */
	io_usleep(25001);	/* P11=25ms P10=400ns */

	pic24_nvmcon_write_completion(0x8000, 0x0000);

	pic24_standby();

	/* PROGRAM FBOOT */
	pic24_write_fboot();
}

/*
 * BLANK PIC24F KA/KM/KL
 *
 * DS39919C-page 11
 */
#undef CALIBKA
void
pic24_PIC24FKA_erase(void)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();
#ifdef CALIBKA
	/* ERASE EXEC & CALIBRATION */
	pic24_six(0x2405AA, 0);	/* MOV #0x405A, W10   */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON    */
	pic24_six(0x200800, 0);	/* MOV #<PAGEVAL>, W0 */
	pic24_six(0x880190, 0);	/* MOV W0, TBLPAG     */
	pic24_six(0x200000, 0);	/* MOV #0x0000, W0    */
	pic24_six(0xBB0800, 2);	/* TBLWTL W0, [W0]    */

	pic24_nvmcon_write_internal();
#endif
	/* 4064 ERASE CODE, DATA, PROTECTION & CONFIG */
	pic24_six(0x24064A, 0);	/* MOV #0x4064, W10   */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON    */
	pic24_six(0x200000, 0);	/* MOV #<PAGEVAL>, W0 */
	pic24_six(0x880190, 0);	/* MOV W0, TBLPAG     */
	pic24_six(0x200000, 0);	/* MOV #0x0000, W0    */
	pic24_six(0xBB0800, 2);	/* TBLWTL W0, [W0]    */

	pic24_nvmcon_write_internal();
#ifdef CALIBKA
	/* RESTORE CALIBRATION WORDS */
	pic24_restore_calibration();
#endif
	pic24_standby();
}

/*
 * BLANK dsPIC33F/PIC24H
 */
void
pic24_dsPIC33F_PIC24H_erase(void)
{
	/* NVMCON = 0x760 */
	pic24_program_verify();

	/* ERASE CONFIG */
	pic24_write_config_words();

	/* ERASE FUID */
	pic24_write_fuid_words();

	/* 404F ERASE CHIP */
	pic24_six(0x2404FA, 0);	/* MOV #0x404F, W10     */
	pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON      */
	pic24_six(0x200000, 0);	/* MOV #<PAGEVAL>, W0   */
	pic24_six(0x880190, 0);	/* MOV W0, TBLPAG       */
	pic24_six(0xA8E761, 4);	/* BSET NVMCON, #WR     */
	io_usleep(330001);	/* P11=330ms, P10=400ns */

	pic24_nvmcon_write_completion(0x8000, 0x0000);

	/* RESTORE CALIBRATION WORDS */
	pic24_restore_calibration();

	pic24_standby();
}

/*
 * BLANK dsPIC33E/PIC24E
 */
void
pic24_dsPIC33EP_PIC24EP_erase(void)
{
	/* NVMCON = 0x728 */
	pic24_program_verify();

	/* 400F ERASE CODE, EXEC & FUID */
	pic24_six(0x2400FA, 1);	/* MOV #0x400F, W10   */
	pic24_six(0x88394A, 2);	/* MOV W10, NVMCON    */
	pic24_six(0x200551, 0);	/* MOV #0x55, W1      */
	pic24_six(0x883971, 0);	/* MOV W1, NVMKEY     */
	pic24_six(0x200AA1, 0);	/* MOV #0xAA, W1      */
	pic24_six(0x883971, 0);	/* MOV W1, NVMKEY     */
	pic24_six(0xA8E729, 3);	/* BSET NVMCON, #WR   */
	io_usleep(21001);	/* P11=21ms P10=400ns */

	pic24_nvmcon_write_completion(0xFFFF, 0x400F);

	pic24_standby();
}

/*
 * BLANK dsPIC33EV/dsPIC33EPXXGS5XX/dsPIC33EPGS2XX
 *
 *  DS70005160C-page 65 ERASE OPERATIONS WITH DUAL PARTITION FLASH
 */
void
pic24_dsPIC33EV_GS_erase(void)
{
	/* NVMCON = 0x728 */
	pic24_program_verify();

	/* 400E ERASE CODE */
	pic24_six(0x2400EA, 1);	/* MOV #0x400E, W10   */
	pic24_six(0x88394A, 2);	/* MOV W10, NVMCON    */
	pic24_six(0x200551, 0);	/* MOV #0x55, W1      */
	pic24_six(0x883971, 0);	/* MOV W1, NVMKEY     */
	pic24_six(0x200AA1, 0);	/* MOV #0xAA, W1      */
	pic24_six(0x883971, 0);	/* MOV W1, NVMKEY     */
	pic24_six(0xA8E729, 3);	/* BSET NVMCON, #WR   */
	io_usleep(24001);	/* P11=24ms P10=400ns */

	pic24_nvmcon_write_completion(0xFFFF, 0x400E);

	pic24_standby();

	/* PROGRAM FBOOT */
	pic24_write_fboot();
}

/*
 * BLANK DEVICE
 *
 * DISABLE PROTECTION AND BULK ERASE
 */
void
pic24_bulk_erase(void)
{
	/* RESET CONFIG & FUID FOR ERASURE */
	memset(&pic24_conf.fuid, -1, sizeof(uint32_t) * PIC24_FUID_MAX);
	memset(&pic24_conf.config[PIC24_PART_LOWER], -1, sizeof(uint32_t) * PIC24_CONFIG_MAX);
	memset(&pic24_conf.config[PIC24_PART_UPPER], -1, sizeof(uint32_t) * PIC24_CONFIG_MAX);

	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		pic24_dsPIC30F_erase(200000);
		break;
	case DS39768D:	  /* PIC24FJ          */
	case DS39934B:	  /* PIC24FJXXGA1/GB0 */
	case DS30009907C: /* PIC24FJXXGA1/GB1 */
		pic24_PIC24FJ_erase(0x880190);
		break;
	case DS30000510E: /* PIC24FJXXGA2/GB2 */
	case DS39970E:    /* PIC24FJXXDA1/DA2/GB2/GA3/GC0 */
		pic24_PIC24FJ_erase(0x8802A0);
		break;
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		pic24_PIC24FJ_GAB_erase();
		break;
	case DS70659C: /* dsPIC33FJ         */
	case DS75012B: /* PIC24FJXXMC       */
		pic24_dsPIC33FJ_erase();
		break;
	case DS39919C: /* PIC24F KA         */
	case DS30625D: /* PIC24F KM/KL      */
		pic24_PIC24FKA_erase();
		break;
	case DS70152H: /* dsPIC33F/PIC24H   */
		pic24_dsPIC33F_PIC24H_erase();
		break;
	case DS70663D: /* dsPIC33EP/PIC24EP */
		pic24_dsPIC33EP_PIC24EP_erase();
		break;
	case DS70005137F: /* dsPIC33EV        */
	case DS70005160C: /* dsPIC33EPXXGS5   */
	case DS70005192B: /* dsPIC33EPXXGS2   */
	case DS70005256A: /* dsPIC33EPXXGS7/8 */
		pic24_dsPIC33EV_GS_erase();
		break;
	default:printf("%s: information: unimplemented\n", __func__);
		break;
	}
}

/*
 * ROW ERASE AT REGION
 */
void
pic24_row_erase(uint32_t row, uint32_t nrows)
{
	printf("%s: information: unimplemented\n", __func__);
}

/*
 * ERASE ROW AT ADDRESS
 *
 *  TODO ADD ERASE ROW SIZE TO DEVICE TABLE AND IMPLEMENT NROWS
 */
void
pic24_erase_row(uint32_t address, uint32_t nrows)
{
	uint32_t w3_nvmadr    = 0x880003 | (PIC24_NVMADR  << 3);
	uint32_t w4_nvmadru   = 0x880004 | (PIC24_NVMADRU << 3);
	uint32_t w10_nvmcon   = 0x88000A | (PIC24_NVMCON  << 3);
	uint32_t w1_nvmkey    = 0x880001 | (PIC24_NVMKEY  << 3);
	uint32_t bs_nvmcon_wr = 0xA8E001 | PIC24_NVMCON;

	pic24_six(0x200FAC, 0); /* MOV #0xFA, W12  */
	pic24_six(0x8802AC, 0); /* MOV W12, TBLPAG */

	do {
		uint32_t w3 = 0x200003 | ((address & 0x00FFFF) << 4);
		uint32_t w4 = 0x200004 | ((address & 0xFF0000) >> 12);

		pic24_six(w3, 0);          /* MOV #DestinationAddress<15:0>, W3  */
		pic24_six(w4, 0);          /* MOV #DestinationAddress<23:16>, W4 */
		pic24_six(w3_nvmadr, 0);   /* MOV W3, NVMADR   */
		pic24_six(w4_nvmadru, 0);  /* MOV W4, NVMADRU  */

		pic24_six(0x24003A, 1);    /* MOV #0x4003, W10 */
		pic24_six(w10_nvmcon, 2);  /* MOV W10, NVMCON  */

		pic24_six(0x200551, 0);    /* MOV #0x55, W1    */
		pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY   */
		pic24_six(0x200AA1, 0);    /* MOV #0xAA, W1    */
		pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY   */
		pic24_six(bs_nvmcon_wr, 3);/* BSET NVMCON, #WR */

		io_usleep(5000);           /* P12=? P10=400ns  */

		pic24_nvmcon_write_completion(0xFFFF, 0x4003);
	}
	while (0);
}

/*****************************************************************************
 *
 * Read block data
 *
 *****************************************************************************/

/*
 * GET DEVICE INDEX
 */
static uint32_t
pic24_get_dev(void)
{
	uint32_t dev = 0;

	/* Device id/rev */
	pic24_set_read_pointer(PIC24_DEVID);
	pic24_conf.deviceid = pic24_table_read16_post_increment();
	pic24_conf.revision = pic24_table_read16_post_increment();

	while (pic24_map[dev].deviceid) {
		if (pic24_map[dev].deviceid == pic24_conf.deviceid)
			break;
		++dev;
	}
	return dev;
}

/*
 * GET CONFIGURATION
 */
int
pic24_read_config_memory(void)
{
	uint32_t dev = 0;

	/* NULL device */
	pic24_index = PIC24_SIZE;

	/* Reset configuraton */
	memset(&pic24_conf, 0, sizeof(pic24_conf));

	/* Device selected by user */
	if (p.devicename[0]) {
		for (dev = 0; pic24_map[dev].deviceid; ++dev) {
			if (strcasecmp(pic24_map[dev].devicename, p.devicename) == 0) {
				/* Device recognised */
				pic24_index = dev;
				break;
			}
		}
		if (pic24_map[dev].deviceid == 0) {
			printf("%s: information: unknown device: [%s]\n", __func__, p.devicename);
			return -1;
		}
	}

	/* Device id/rev */
	switch (pic24_map[pic24_index].datasheet) {
	default:
		pic24_conf.tblpag  = 0x0032;
		pic24_conf.nvmcon  = 0x0760;
		pic24_conf.visi    = 0x0784;
		pic24_conf.tblnop  = 2 + 1 - 1;	/* TBLRD: 2 CYCLES + 1 STALL */
		pic24_conf.gotonop = 2 + 0 - 1;	/* GOTO:  2 CYCLES + 0 STALL */
		pic24_program_verify();
		pic24_index = pic24_get_dev();
#if 0
		if (!pic24_map[pic24_index].deviceid) {
			pic24_standby();
			pic24_conf.tblpag  = 0x0054;
			pic24_program_verify();
			pic24_index = pic24_get_dev();
		}
#endif
		break;
	case DS70663D:		/* EP GP/MC   */
	case DS70005137F:	/* EV GM      */
	case DS70005160C:	/* EP GS5     */
	case DS70005192B:	/* EP GS2     */
	case DS70005256A:	/* EP GS7 GS8 */
		pic24_conf.tblpag  = 0x0054;
		pic24_conf.nvmcon  = 0x0728;
		pic24_conf.visi    = 0x0F88;
		pic24_conf.tblnop  = 5 + 1 - 1;	/* TBLRD: 5 CYCLES + 1 STALL */
		pic24_conf.gotonop = 4 + 0 - 1;	/* GOTO:  4 CYCLES + 0 STALL */
		pic24_program_verify();
		pic24_index = pic24_get_dev();
		break;
	case DS39970E:	  /* PIC24FJXXDA1/DA2/GB2/GA3/GC0 */
	case DS30000510E: /* PIC24FJXXGA2/GB2 */
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		pic24_conf.tblpag  = 0x0054;
		pic24_conf.nvmcon  = 0x0760;
		pic24_conf.visi    = 0x0784;
		pic24_conf.tblnop  = 2 + 1 - 1; /* TBLRD: 2 CYCLES + 1 STALL */
		pic24_conf.gotonop = 2 + 0 - 1; /* GOTO:  2 CYCLES + 0 STALL */
		pic24_program_verify();
		pic24_index = pic24_get_dev();
		break;
	}
	if (!pic24_map[pic24_index].deviceid) {
		if (pic24_conf.deviceid == 0x0000 || pic24_conf.deviceid == 0xFFFF) {
			printf("%s: information: %s\n",
				__func__, io_fault(pic24_conf.deviceid));
		} else {
			printf("%s: information: device unknown: [%06X] [%06X]\n",
				__func__, pic24_conf.deviceid, pic24_conf.revision);
		}
		pic24_standby();
		return -1;
	}
#if 0
	/* Device family */
	dev = 0;
	while (pic24_tab[dev].datasheet) {
		if (pic24_tab[dev].datasheet == pic24_map[pic24_index].datasheet) {
			pic_pe_lookup(pic24_conf.pepath, pic24_tab[dev].filename);
			break;
		}
		++dev;
	}
#endif
	/* FBOOT */
	if (pic24_supports_partition()) {
		pic24_set_read_pointer(pic24_map[pic24_index].fbootaddr);
		pic24_conf.fboot = pic24_table_read24_post_increment();
	}

	/* CONFIG */
	if (pic24_has_partition()) {
		/* CONFIG lower */
		pic24_set_read_pointer(pic24_map[pic24_index].configlower);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i)
			pic24_conf.config[PIC24_PART_LOWER][i] = pic24_table_read16_post_increment();

		/* FBTSEQ lower */
		pic24_set_read_pointer(pic24_map[pic24_index].configlower | PIC24_FBTSEQ);
		pic24_conf.config[PIC24_PART_LOWER][PIC24_FBTSEQ] = pic24_table_read24_post_increment();

		/* CONFIG upper */
		pic24_set_read_pointer(pic24_map[pic24_index].configlower | PIC24_CODE_UPPER);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i)
			pic24_conf.config[PIC24_PART_UPPER][i] = pic24_table_read16_post_increment();

		/* FBTSEQ upper */
		pic24_set_read_pointer(pic24_map[pic24_index].configlower | PIC24_CODE_UPPER | PIC24_FBTSEQ);
		pic24_conf.config[PIC24_PART_UPPER][PIC24_FBTSEQ] = pic24_table_read24_post_increment();
	} else {
		/* CONFIG */
		pic24_set_read_pointer(pic24_map[pic24_index].configaddr);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i)
			pic24_conf.config[PIC24_CONFIG][i] = pic24_table_read24_post_increment();

		/* FBTSEQ */
		if (pic24_supports_partition()) {
			pic24_set_read_pointer(pic24_map[pic24_index].configaddr | PIC24_FBTSEQ);
			pic24_conf.config[PIC24_CONFIG][PIC24_FBTSEQ] = pic24_table_read24_post_increment();
		}
	}

	/* FUID */
	if (pic24_map[pic24_index].fuidaddr) {
		pic24_set_read_pointer(pic24_map[pic24_index].fuidaddr);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nfuid; ++i)
			pic24_conf.fuid[i] = pic24_table_read24_post_increment();
	}

	/* Calibration */
	if (pic24_map[pic24_index].calibaddr) {
		pic24_set_read_pointer(pic24_map[pic24_index].calibaddr);
		for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i)
			pic24_conf.calib[i] = pic24_table_read24_post_increment();
	}

	pic24_standby();

	return 0;
}

/*
 * GET PROGRAM COUNT
 *
 *  RETURN CODE PARTITION COUNT
 */
uint32_t
pic24_get_program_count(void)
{
	if (pic24_has_partition()) {
		return 2;
	}
	return 1;
}

/*
 * GET PROGRAM FLASH SIZE
 *
 *  RETURN CODE SIZE IN WORDS
 */
uint32_t
pic24_get_program_size(uint32_t *addr, uint32_t partition)
{
	if (pic24_has_partition()) {
		if (partition == PIC_PART2)
			*addr = PIC24_CODE_UPPER;
		else
			*addr = PIC24_CODE_LOWER;

		return pic24_map[pic24_index].flash >> 1;
	}
	*addr = PIC24_CODE_LOWER;

	return pic24_map[pic24_index].flash;
}

/*
 * GET DATA EEPROM/FLASH SIZE
 *
 *  RETURN DATA SIZE IN BYTES
 */
uint32_t
pic24_get_data_size(uint32_t *addr)
{
	if (pic24_map[pic24_index].eeprom) {
		*addr = PIC24_EXEC_LOW - pic24_map[pic24_index].eeprom;
	} else {
		*addr = 0;
	}
	return pic24_map[pic24_index].eeprom;
}

/*
 * GET EXECUTIVE SIZE
 *
 *  RETURN EXEC SIZE IN WORDS
 */
uint32_t
pic24_get_executive_size(uint32_t *addr)
{
	uint32_t size = 0;

	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		*addr = PIC24_EXEC_LOW;
		size = dsPIC30F_EXEC_HIGH - *addr;
		break;
	case DS70663D:    /* dsPIC33EP/PIC24EP */
	case DS70005137F: /* dsPIC33EV         */
	case DS70005160C: /* dsPIC33EPXXGS5XX  */
	case DS70005192B: /* dsPIC33EPXXGS2XX  */
	case DS70005256A: /* dsPIC33EPXXGS7/8  */
		*addr = PIC24_EXEC_LOW;
		size = dsPIC33E_EXEC_HIGH - *addr;
		break;
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		*addr = PIC24_GAB_EXEC_LOW;
		size = PIC24_GA7_EXEC_HIGH - *addr;
		break;
	default:*addr = PIC24_EXEC_LOW;
		size = PIC24_EXEC_HIGH - *addr;
		break;
	}

	return (size >> 1);
}

/*
 * READ PROGRAM FLASH MEMORY BLOCK ADDR .. ADDR + SIZE
 *
 *  RETURN ADDR
 */
#define GOTO200 (32)
uint32_t
pic24_read_program_memory_block(uint32_t *data, uint32_t addr, uint32_t size)
{
	if (size & 1) {
		printf("%s: fatal error: block size (%u) uneven\n", __func__, size);
		io_exit(EX_SOFTWARE); /* Panic */
	}

	pic24_program_verify();

	pic24_set_read_pointer(addr);
	uint32_t pc = 0;
	for (uint32_t i = 0; i < size;) {
		if ((pc++ % GOTO200) == 0)
			pic24_goto200();
		uint32_t word0, word1;
		pic24_table_read48_post_increment(&word0, &word1);
		data[i++] = (uint32_t)word0;
		data[i++] = (uint32_t)word1;
#ifdef DEBUG
		char d[STRLEN];
		int rc = snprintf(d, STRLEN, "%s() addr=%06X size=%d", __func__, addr, size);
		if (rc < 0) {
			printf("%s: fatal error: snprintf failed\n", __func__);
			io_exit(EX_OSERR); /* Panic */
		} else if (rc >= STRLEN) {
			printf("%s: fatal error: snprintf overrun\n", __func__);
			io_exit(EX_SOFTWARE); /* Panic */
		}
		printf("%s: PCL=%04X\n", d, pic24_readreg(_PCL));
#endif
		addr += 4;
		if ((addr % 0x0010000) == 0)
			pic24_set_read_pointer(addr);
	}

	pic24_standby();

	return addr;
}

/*
 * READ DATA EEPROM/FLASH MEMORY BLOCK ADDR .. ADDR + SIZE
 *
 *  RETURN ADDR
 */
uint32_t
pic24_read_data_memory_block(uint16_t *data, uint32_t addr, uint16_t size)
{
	pic24_program_verify();

	pic24_set_read_pointer(addr);
	uint32_t pc = 0;
	for (uint32_t i = 0; i < size;) {
		if ((pc++ % GOTO200) == 0)
			pic24_goto200();
		data[i++] = pic24_table_read16_post_increment();
#ifdef DEBUG
		char d[STRLEN];
		int rc;
		rc = snprintf(d, STRLEN, "%s() addr=%06X size=%d", __func__, addr, size);
		if (rc < 0) {
			printf("%s: fatal error: snprintf failed\n", __func__);
			io_exit(EX_OSERR); /* Panic */
		} else if (rc >= STRLEN) {
			printf("%s: fatal error: snprintf overrun\n", __func__);
			io_exit(EX_SOFTWARE); /* Panic */
		}
		printf("%s: PCL=%04X\n", d, pic24_readreg(_PCL));
#endif
	}

	pic24_standby();

	return addr;
}

/*****************************************************************************
 *
 * Program Code
 *
 *****************************************************************************/

/*
 * WRITE PROGRAM INIT
 *
 *  SETUP FOR PROGRAM OR DATA WRITING
 */
void
pic24_write_program_init(void)
{
	/* RESET CONFIG & FUID FOR PROGRAMMING */
	memset(&pic24_conf.fuid, 0, sizeof(uint32_t) * PIC24_FUID_MAX);
	memset(&pic24_conf.config[PIC24_PART_LOWER], 0, sizeof(uint32_t) * PIC24_CONFIG_MAX);
	memset(&pic24_conf.config[PIC24_PART_UPPER], 0, sizeof(uint32_t) * PIC24_CONFIG_MAX);

	switch (pic24_map[pic24_index].datasheet) {
	case DS39768D:    /* PIC24FJ                      */
	case DS39970E:    /* PIC24FJXXDA1/DA2/GB2/GA3/GC0 */
	case DS70152H:    /* dsPIC33F/PIC24H              */
	case DS39934B:    /* PIC24FJXXGA1/GB0             */
	case DS30000510E: /* PIC24FJXXGA2/GB2             */
	case DS30009907C: /* PIC24FJXXGA1/GB1             */
		/* NVMCON = 0x760 */
		pic24_six(0x24001A, 0);	/* MOV #0x4001, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
		break;
	case DS70659C: /* dsPIC33FJ    */
	case DS75012B: /* PIC24FJXXMC  */
		/* NVMCON = 0x760 */
		pic24_six(0x24003A, 0);	/* MOV #0x4003, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
		break;
	case DS39919C: /* PIC24F KA    */
	case DS30625D: /* PIC24F KM/KL */
		/* NVMCON = 0x760 */
		pic24_six(0x24004A, 0);	/* MOV #0x4004, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
		break;
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		/* NVMCON = 0x760 */
		pic24_six(0x240020, 0); /* MOV #0x4002, W0  */
		pic24_six(0x883B00, 0);	/* MOV W0, NVMCON   */
		/* Follow through */
	case DS70663D:	  /* dsPIC33EP/PIC24EP */
	case DS70005137F: /* dsPIC33EV         */
	case DS70005160C: /* dsPIC33EPXXGS5    */
	case DS70005192B: /* dsPIC33EPXXGS2    */
	case DS70005256A: /* dsPIC33EPXXGS7/8  */
		/* DWORD writing */
		pic24_six(0x200FAC, 0);	/* MOV #0xFA, W12   */
		pic24_six(0x8802AC, 0);	/* MOV W12, TBLPAG  */
	default:break;
	}
}

/*
 * WRITE 24-BIT PROGRAM BUFFER TO 24-BIT PROGRAM FLASH
 */
void
pic24_write_program(uint32_t address, uint32_t *buffer, uint32_t buffer_size)
{
#if 0
	printf("%s() address=%06X buffer_size=%d\n", __func__, address, buffer_size);
	pic24_dumphexcode(address, buffer_size, buffer);
#else
	switch (pic24_map[pic24_index].datasheet) {
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		pic24_set_write_pointer_nvmadr(address);
		break;
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		/* NVMCON = 0x760 */
		pic24_six(0x24001A, 0);	/* MOV #0x4001, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
	default:
		pic24_set_write_pointer(address);
		break;
	}

	for (uint32_t i = 0; i < buffer_size; i += 4) {
		uint16_t lsw0 = buffer[i + 0];
		uint16_t lsw1 = buffer[i + 1];
		uint16_t lsw2 = buffer[i + 2];
		uint16_t lsw3 = buffer[i + 3];
		uint8_t  msb0 = buffer[i + 0] >> 16;
		uint8_t  msb1 = buffer[i + 1] >> 16;
		uint8_t  msb2 = buffer[i + 2] >> 16;
		uint8_t  msb3 = buffer[i + 3] >> 16;

		uint32_t w0 = 0x200000 | (lsw0 << 4);
		uint32_t w1 = 0x200001 | (msb1 << 12) | (msb0 << 4);
		uint32_t w2 = 0x200002 | (lsw1 << 4);
		uint32_t w3 = 0x200003 | (lsw2 << 4);
		uint32_t w4 = 0x200004 | (msb3 << 12) | (msb2 << 4);
		uint32_t w5 = 0x200005 | (lsw3 << 4);

		/* LOAD W0:W5 WITH 4 INSTRUCTIONS */
		pic24_six(w0, 0);	/* MOV #<LSW0>, W0      */
		pic24_six(w1, 0);	/* MOV #<MSB1:MSB0>, W1 */
		pic24_six(w2, 0);	/* MOV #<LSW1>, W2      */
		pic24_six(w3, 0);	/* MOV #<LSW2>, W3      */
		pic24_six(w4, 0);	/* MOV #<MSB3:MSB2>, W4 */
		pic24_six(w5, 0);	/* MOV #<LSW3>, W5      */

		/* SET THE READ POINTER AND LOAD LATCHES */
		pic24_six(0xEB0300, 1);/* CLR W6                  */
		pic24_six(0xBB0BB6, 2);/* TBLWTL   [W6++], [W7]   */
		pic24_six(0xBBDBB6, 2);/* TBLWTH.B [W6++], [W7++] */
		pic24_six(0xBBEBB6, 2);/* TBLWTH.B [W6++], [++W7] */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
		pic24_six(0xBB0BB6, 2);/* TBLWTL   [W6++], [W7]   */
		pic24_six(0xBBDBB6, 2);/* TBLWTH.B [W6++], [W7++] */
		pic24_six(0xBBEBB6, 2);/* TBLWTH.B [W6++], [++W7] */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
	}

	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		/* DS70102H-page 90 dsPIC30F4013 TPROG P12(2ms MAX) */
		/* DS70284B-page 44 dsPIC30F1010 TPROG P18(4ms MAX) */
		pic24_nvmcon_write_external(4000);
		break;
	case DS30010073A: /* PIC24FJXXGA4/GB4 */
	case DS30010057F: /* PIC24FJXXGA6/GB6 */
	case DS30010102A: /* PIC24FJXXGA7     */
		pic24_nvmcon_write_initiate_GAB();
		/* Follow through */
	default:pic24_nvmcon_write_internal();
		break;
	}
#endif
}

/*
 * WRITE 16-BIT DATA BUFFER TO 16-BIT DATA EEPROM
 *
 *  TBLPAG = 0x7F
 */
void
pic24_write_data(uint32_t address, uint32_t *buffer, uint32_t buffer_size)
{
	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F */
		/* NVMCON = 0x760 */
		pic24_six(0x24005A, 0);	/* MOV #0x4005, W10  */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON   */
	default:break;
	}
	pic24_set_write_pointer(address);
	for (uint32_t i = 0; i < buffer_size; i += 4) {
		uint32_t w0 = 0x200000 | ((0xFFFF & buffer[i + 0]) << 4);
		uint32_t w1 = 0x200001 | ((0xFFFF & buffer[i + 1]) << 4);
		uint32_t w2 = 0x200002 | ((0xFFFF & buffer[i + 2]) << 4);
		uint32_t w3 = 0x200003 | ((0xFFFF & buffer[i + 3]) << 4);

		/* LOAD W0:W3 WITH 4 DATA WORDS */
		pic24_six(w0, 0);	/* MOV #<WORD0>, W0 */
		pic24_six(w1, 0);	/* MOV #<WORD1>, W1 */
		pic24_six(w2, 0);	/* MOV #<WORD2>, W2 */
		pic24_six(w3, 0); 	/* MOV #<WORD3>, W3 */

		/* SET THE READ POINTER AND LOAD LATCHES */
		pic24_six(0xEB0300, 1);/* CLR W6                  */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
		pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */
	}	

	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F */
		/* DS70102H-page 90 dsPIC30F4013 TPROG P12(2ms MAX) */
		pic24_nvmcon_write_external(2000);
		break;
	default:pic24_nvmcon_write_internal();
		break;
	}
}

/*
 * WRITE 24-BIT PROGRAM WORD TO 24-BIT PROGRAM FLASH
 *
 *  TBLPAG = 0x00
 */
void
pic24_write_program_word24(uint32_t address, uint32_t word)
{
	uint32_t w2 = 0x200002 | ((0x0000FFFF & address) << 4);
	uint32_t w5 = 0x200005 | ((0x0000FFFF & word)    << 4);
	uint32_t w6 = 0x200006 | ((0x00FF0000 & word)    >> 12);

	pic24_six(w2, 0);	/* MOV #<DestinationAddress15:0>, W2 */
	pic24_six(w5, 0);	/* MOV #<LOW_WORD>, W5               */
	pic24_six(w6, 1);	/* MOV #<HIGH_BYTE>, W6              */
	pic24_six(0xBB0905, 2);	/* TBLWTL W5, [W2]                   */
	pic24_six(0xBB9906, 3);	/* TBLWTH.B W6, [W2++]               */

	pic24_nvmcon_write_internal();
}

/*
 * WRITE 24-BIT x 2 PROGRAM WORDS TO 24-BIT PROGRAM FLASH
 *
 *  TBLPAG = 0xFA
 */
void
pic24_write_program_word48(uint32_t address, uint32_t word0, uint32_t word1)
{
	uint32_t w3_nvmadr    = 0x880003 | (PIC24_NVMADR  << 3);
	uint32_t w4_nvmadru   = 0x880004 | (PIC24_NVMADRU << 3);
	uint32_t w10_nvmcon   = 0x88000A | (PIC24_NVMCON  << 3);
	uint32_t w1_nvmkey    = 0x880001 | (PIC24_NVMKEY  << 3);
	uint32_t bs_nvmcon_wr = 0xA8E001 | PIC24_NVMCON;

	uint16_t lsw0 = word0;
	uint16_t lsw1 = word1;
	uint8_t  msb0 = word0 >> 16;
	uint8_t  msb1 = word1 >> 16;

	uint32_t w0 = 0x200000 | (lsw0 << 4);
	uint32_t w1 = 0x200001 | (msb1 << 12) | (msb0 << 4);
	uint32_t w2 = 0x200002 | (lsw1 << 4);

	/* LOAD W0:W2 WITH 2 INSTRUCTIONS */
	pic24_six(w0, 0);	/* MOV #<LSW0>, W0      */
	pic24_six(w1, 0);	/* MOV #<MSB1:MSB0>, W1 */
	pic24_six(w2, 0);	/* MOV #<LSW1>, W2      */

	/* SET THE READ POINTER (W6) AND WRITE POINTER (W7) AND LOAD THE WRITE LATCHES */
	pic24_six(0xEB0300, 1);/* CLR W6                  */
	pic24_six(0xEB0380, 1);/* CLR W7                  */
	pic24_six(0xBB0BB6, 2);/* TBLWTL   [W6++], [W7]   */
	pic24_six(0xBBDBB6, 2);/* TBLWTH.B [W6++], [W7++] */
	pic24_six(0xBBEBB6, 2);/* TBLWTH.B [W6++], [++W7] */
	pic24_six(0xBB1BB6, 2);/* TBLWTL   [W6++], [W7++] */

	uint32_t w3 = 0x200003 | ((address & 0x00FFFF) << 4);
	uint32_t w4 = 0x200004 | ((address & 0xFF0000) >> 12);

	/* SET THE NVMADRU/NVMADR REGISTER-PAIR TO POINT TO THE CORRECT ADDRESS */
	pic24_six(w3, 0);       /* MOV #DestinationAddress<15:0>, W3  */
	pic24_six(w4, 0);       /* MOV #DestinationAddress<23:16>, W4 */
	pic24_six(w3_nvmadr,  0);/* MOV W3, NVMADR  */
	pic24_six(w4_nvmadru, 0);/* MOV W4, NVMADRU */

	/* SET THE NVMCON TO PROGRAM TWO INSTRUCTION WORDS */
	pic24_six(0x24001A, 1);  /* MOV 0x4001, W10 */
	pic24_six(w10_nvmcon, 2);/* MOV W10, NVMCON */

	/* INITIATE THE WRITE CYCLE */
	pic24_six(0x200551, 0);    /* MOV #0x55, W1     */
	pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY    */
	pic24_six(0x200AA1, 0);    /* MOV #0xAA, W1     */
	pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY    */
	pic24_six(bs_nvmcon_wr, 3);/* BSET NVMCON, #WR  */

	/* DS700000657H TWW(53.8us) or TWW(54.4us)     */
	io_usleep(55); /* P13=TWW P10=400ns */

	pic24_nvmcon_write_completion(0xFFFF, 0x4001);
}

/*
 * WRITE 16-BIT DATA WORD TO 16-BIT DATA EEPROM
 *
 *  TBLPAG = 0x7F
 */
void
pic24_write_data_word16(uint32_t address, uint32_t word)
{
	uint32_t w2 = 0x200002 | ((0x0000FFFF & address) << 4);
	uint32_t w5 = 0x200005 | ((0x0000FFFF & word)    << 4);

	pic24_six(w2, 0);	/* MOV #<DestinationAddress15:0>, W2 */
	pic24_six(w5, 0);	/* MOV #<DATA_WORD_VALUE>, W5        */
	pic24_six(0xBB0905, 2);	/* TBLWTL W5, [W2]                   */

	pic24_nvmcon_write_internal();
}

/*
 * WRITE PANEL TO 24-BIT PROGRAM FLASH OR 16-BIT DATA EEPROM
 */
void
pic24_write_panel(uint32_t region, uint32_t address, uint32_t *panel, uint32_t panel_size)
{
	switch (region) {
	case PIC_REGIONCODE:
	case PIC_REGIONPART:
	case PIC_REGIONEXEC:
		if (panel_size == 1) {		/* dsPIC33FJ/PIC24FJXXMC WORD WRITING */
			pic24_write_program_word24(address << 1, panel[0]);
		} else if (panel_size == 2) {	/* dsPIC33EP/PIC24EP/dsPIC33EV DWORD WRITING */
			pic24_write_program_word48(address << 1, panel[0], panel[1]);
		} else {			/* PANEL WRITING */
			pic24_write_program(address << 1, panel, panel_size);
		}
		break;
	case PIC_REGIONDATA:
		if (panel_size == 1) {		/* PIC24F KA KM/KL WORD WRITING */
			pic24_write_data_word16(address << 1, panel[0]);
		} else {			/* PANEL WRITING */
			pic24_write_data(address << 1, panel, panel_size);
		}
		break;
	default:printf("%s: warning: region unsupported [%d]\n", __func__, region);
		break;
	}
}

/*
 * HAS FUID
 *
 *  RETURN TRUE IF FUID EXISTS
 */
static inline uint32_t
pic24_has_fuid(void)
{
	return pic24_map[pic24_index].fuidaddr;
}

/*
 * HAS CONFIG SPACE
 *
 *  RETURN TRUE IF CONFIG IN CONFIG SPACE
 */
static inline uint32_t
pic24_has_config(void)
{
	if (pic24_map[pic24_index].configaddr == PIC24_CONFIG_ADDR)
		return PIC24_CONFIG_ADDR; /* CONFIG SPACE */

	return 0;
}

/*
 * CONFIG INIT
 *
 *  SETUP FOR CONFIG OR FUID WRITING IN THE CONFIG SPACE
 */
void
pic24_write_config_init1(void)
{
	switch (pic24_map[pic24_index].datasheet) {
	case DS39919C: /* PIC24F KA    */
	case DS30625D: /* PIC24F KM/KL */
		/* NVMCON = 0x760 */
		pic24_six(0x24004A, 0);	/* MOV #0x4004, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
		break;
	case DS70152H: /* dsPIC33F/PIC24H */
		/* NVMCON = 0x760  */
		pic24_six(0x24000A, 0);	/* MOV #0x4000, W10 */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON  */
	default:break;
	}
}

void
pic24_write_config_init2(void)
{
	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		/* NVMCON = 0x760 */
		pic24_six(0x24008A, 0);	/* MOV #0x4008, W10    */
		pic24_six(0x883B0A, 0);	/* MOV W10, NVMCON     */
	default:break;
	}
}

void
pic24_write_config_word(uint32_t address, uint16_t word)
{
	uint32_t w6 = 0x200006 | (word << 4);

	/* INITIALISE TBLPAG AND THE WRITE POINTER (W7) FOR TBLWT INSTRUCTION */
	pic24_set_write_pointer(address);	/* TBLPAG=0xF8 W7=<15:0> */

	/* LOAD THE CONFIGURATION DATA REGISTER TO W6 */
	pic24_six(w6, 1);			/* MOV #<VALUE>, W6      */

	/* WRITE THE CONFIGURATION REGISTER DATA TO THE WRITE LATCH AND INC. WRITE POINTER */
	pic24_six(0xBB1B86, 2);			/* TBLWTL W6, [W7++]     */

	switch (pic24_map[pic24_index].datasheet) {
	case DS70102H: /* dsPIC30F      */
	case DS70284B: /* dsPIC30F SMPS */
		/* DS70102H-page 90 dsPIC30F4013 TPROG P12(2ms MAX) */
		/* DS70284B-page 44 dsPIC30F1010 TPROG P18(4ms MAX) */
		/* UNLOCK NVMCON AND WRITE */
		pic24_nvmcon_write_external(4000);
		break;
	default:/* WRITE */
		pic24_nvmcon_write_internal();
		break;
	}
}

/*
 * WRITE CONFIG WORDS IN THE CONFIG SPACE
 */
uint32_t
pic24_write_config_words(void)
{
	if (pic24_has_config() == 0)
		return 0; /* Not supported here */

	pic24_write_config_init1();
	uint32_t n = 0;
	for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
		if (pic24_conf.config[PIC24_PART_UPPER][i] & PIC24_CONFIG_PENDING) {
			uint32_t address = pic24_map[pic24_index].configaddr + 2 * i;
			pic24_write_config_init2();
			pic24_write_config_word(address, pic24_conf.config[PIC24_PART_UPPER][i]);
			n++;
		}
	}

	return n;
}

/*
 * WRITE FUID WORDS IN THE CONFIG SPACE
 */
uint32_t
pic24_write_fuid_words(void)
{
	if (pic24_has_fuid() == 0)
		return 0; /* Not supported here */

	pic24_write_config_init1();
	uint32_t n = 0;
	for (uint32_t i = 0; i < pic24_map[pic24_index].nfuid; ++i) {
		if (pic24_conf.fuid[i] & PIC24_CONFIG_PENDING) {
			uint32_t address = pic24_map[pic24_index].fuidaddr + 2 * i;
			pic24_write_config_init2();
			pic24_write_config_word(address, pic24_conf.fuid[i]);
			n++;
		}
	}

	return n;
}

/*
 * WRITE CONFIG & FUID WORDS IN THE CONFIG SPACE
 */
uint32_t
pic24_write_config(void)
{
	if (pic24_has_fuid() || pic24_has_config()) {
		uint32_t n = 0;

		pic24_program_verify();

		if (pic24_map[pic24_index].fuidaddr)
			n += pic24_write_fuid_words();

		if (pic24_map[pic24_index].configaddr)
			n += pic24_write_config_words();

		pic24_standby();

		return n;
	}
	return 0;
}

/*
 * PROGRAM FBOOT DUAL PARTITION
 */
static void
pic24_write_fboot_dual_partition(void)
{
	uint32_t w10_nvmcon   = 0x88000A | (PIC24_NVMCON  << 3);
	uint32_t w1_nvmkey    = 0x880001 | (PIC24_NVMKEY  << 3);
	uint32_t bs_nvmcon_wr = 0xA8E001 | PIC24_NVMCON;

	pic24_six(0x200FAC, 0);    /* MOV #0xFA, W12      */
	pic24_six(0x8802AC, 0);    /* MOV W12, TBLPAG     */

	pic24_six(0xEB0100, 0);    /* CLR W2              */
	pic24_six(0x2FFFE5, 0);    /* MOV #0xFFFE, W5     */
	pic24_six(0x200FF6, 1);    /* MOV #0x00FF, W6     */
	pic24_six(0xBB0905, 2);    /* TBLWTL W5, [W2]     */
	pic24_six(0xBB9906, 3);    /* TBLWTH.B W6, [W2++] */

	pic24_six(0x24008A, 1);    /* MOV #0x4008, W10    */
	pic24_six(w10_nvmcon, 2);  /* MOV W10, NVMCON     */

	pic24_six(0x200551, 0);    /* MOV #0x55, W1       */
	pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY      */
	pic24_six(0x200AA1, 0);    /* MOV #0xAA, W1       */
	pic24_six(w1_nvmkey, 0);   /* MOV W1, NVMKEY      */
	pic24_six(bs_nvmcon_wr, 3);/* BSET NVMCON, #WR    */

	io_usleep(1000);           /* P13B=? P10=400ns    */

	pic24_nvmcon_write_completion(0xFFFF, 0x4008);
}

/*
 * PROGRAM FBOOT
 */
void
pic24_write_fboot(void)
{
	if (pic24_wants_partition() == 0)
		return;	/* Not wanted */

	pic24_program_verify();

	pic24_write_fboot_dual_partition();

	if (pic24_map[pic24_index].datasheet == DS70005160C) {
		/*
		 * This row erase is needed otherwise a ghost FSIGN
		 * write will occur at 0x402B94 due to an unexplained
		 * hardware bug.
		 */
		pic24_erase_row(0x005400, 1);

		/*
		 * FSIGN needs to written here due to an unexplained
		 * hardware bug.
		 */
		pic24_write_program_word48(0x00AB94, 0xFF7FFF, 0xFFFFFF);
	}

	pic24_standby();
}

/*****************************************************************************
 *
 * Programming
 *
 *****************************************************************************/

/*
 * DETERMINE MEMORY REGION: CODE, PARTITION, DATA, CONFIG, EXEC, OR ID
 *
 *  RETURN PIC_REGIONCODE:
 *	000000 .. FLASH - 2
 *
 *  RETURN PIC_REGIONPART:
 *      400000 .. FLASH - 2
 *
 *  RETURN PIC_REGIONDATA:
 *	7FFXXX .. 7FFFFE
 *
 *  RETURN PIC_REGIONCONFIG:
 *	F80000 .. F8000E
 *	XXXXXX .. XXXXXX
 *
 *  RETURN PIC_REGIONEXEC:
 *	800000 .. 80XXXX
 *
 *  RETURN PIC_REGIONID:
 *	8005C0 .. 8005FE
 */
uint32_t
pic24_getregion(uint32_t address)
{
	/* CODE */
	if (pic24_wants_partition()) {
		/* CODE LOWER */
		if (address < pic24_map[pic24_index].flash) {
			return PIC_REGIONCODE;
		}
		/* CODE UPPER */
		if (address >= PIC24_CODE_UPPER && address < (PIC24_CODE_UPPER + pic24_map[pic24_index].flash)) {
			return PIC_REGIONPART;
		}
	} else {
		/* CODE */
		if (address < (pic24_map[pic24_index].flash << 1)) {
			return PIC_REGIONCODE;
		}
	}
	/* FBOOT */
	if (pic24_supports_partition() && (address == pic24_map[pic24_index].fbootaddr)) {
		return PIC_REGIONBOOT;
	}
	/* EEPROM */
	if (pic24_map[pic24_index].eeprom) {
		uint32_t data_low = PIC24_EXEC_LOW - pic24_map[pic24_index].eeprom;
		if (address >= data_low && address < PIC24_EXEC_LOW) {
			return PIC_REGIONDATA;
		}
	}
	/* EXEC */
	if ((pic24_map[pic24_index].datasheet == DS70102H) ||		/* dsPIC30F      */
		(pic24_map[pic24_index].datasheet == DS70284B)) {	/* dsPIC30F SMPS */
		if (address >= PIC24_EXEC_LOW && address < dsPIC30F_EXEC_HIGH) {
			return PIC_REGIONEXEC;
		}
	} else if ((pic24_map[pic24_index].datasheet == DS70663D) || 	/* dsPIC33EP/PIC24EP */
		(pic24_map[pic24_index].datasheet == DS70005137F) ||	/* dsPIC33EV         */
		(pic24_map[pic24_index].datasheet == DS70005160C) ||	/* dsPIC33EPXXGS5XX  */
		(pic24_map[pic24_index].datasheet == DS70005192B) || 	/* dsPIC33EPXXGS2XX  */
		(pic24_map[pic24_index].datasheet == DS70005256A)) {	/* dsPIC33EPXXGS7/8  */
		if (address >= PIC24_EXEC_LOW && address < dsPIC33E_EXEC_HIGH) {
			return PIC_REGIONEXEC;
		}
	} else if ((pic24_map[pic24_index].datasheet == DS30010073A) || /* PIC24FJXXGA4/GB4 */
		(pic24_map[pic24_index].datasheet == DS30010057F) ||	/* PIC24FJXXGA6/GB6 */
		(pic24_map[pic24_index].datasheet == DS30010102A)) {	/* PIC24FJXXGA7     */
		if (address >= PIC24_GAB_EXEC_LOW && address < PIC24_GA7_EXEC_HIGH) {
			return PIC_REGIONEXEC;
		}
	} else if (address >= PIC24_EXEC_LOW && address < PIC24_EXEC_HIGH) {
		return PIC_REGIONEXEC;
	}
	/* FUID */
	if (pic24_has_fuid()) {
		uint32_t fuid_high = pic24_map[pic24_index].fuidaddr + (pic24_map[pic24_index].nfuid << 1);
		if (address >= pic24_map[pic24_index].fuidaddr && address < fuid_high) {
			return PIC_REGIONID;
		}
	}
	/* CONFIG */
	if (pic24_has_config()) {
		uint32_t config_high = pic24_map[pic24_index].configaddr + (pic24_map[pic24_index].nconfig << 1);
		if (address >= pic24_map[pic24_index].configaddr && address < config_high) {
			return PIC_REGIONCONFIG;
		}
	}
	if (p.f) fprintf(p.f, "%s: warning: address unsupported [%06X]\n",
		__func__, address);
	return PIC_REGIONNOTSUP;
}

/*
 * INIT REGION FOR WRITING (CONFIG DELAYED)
 *
 *  RETURN REGION IF WRITING SUPPORTED
 */
static inline uint32_t
pic24_init_writeregion(uint32_t region)
{
	switch (region) {
	case PIC_REGIONCODE:
	case PIC_REGIONPART:
	case PIC_REGIONEXEC:
		if (pic24_map[pic24_index].codepanelsize == 1) { /* dsPIC33FJ WORD WRITING */
			/* TBLPAG = 0x00 */
			pic24_six(0x200001, 1);/* MOV #0x00, W1  */
			pic24_six(0x880191, 0);/* MOV W1, TBLPAG */
		}
		pic_write_panel(PIC_PANEL_BEGIN, region, pic24_map[pic24_index].codepanelsize);
		return region;
	case PIC_REGIONDATA:
		if (pic24_map[pic24_index].datapanelsize == 1) { /* PIC24F KM/KL WORD WRITING */
			/* TBLPAG = 0x7F */
			pic24_six(0x2007F1, 1); /* MOV #0x7F, W1  */
			pic24_six(0x880191, 0); /* MOV W1, TBLPAG */
		}
		pic_write_panel(PIC_PANEL_BEGIN, region, pic24_map[pic24_index].datapanelsize);
	case PIC_REGIONCONFIG:
	case PIC_REGIONID:
	case PIC_REGIONBOOT:
		return region;
	}
	if (p.f) fprintf(p.f, "%s: warning: region unsupported [%d]\n",
		__func__, region);
	return PIC_REGIONNOTSUP;
}

/*
 * WRITE REGION (CACHE CONFIG & FUID WORDS)
 */
static inline void
pic24_writeregion(uint32_t address, uint32_t region, uint32_t data)
{
	switch (region) {
	case PIC_REGIONCODE:
	case PIC_REGIONPART:
	case PIC_REGIONDATA:
	case PIC_REGIONEXEC:
		pic_write_panel(PIC_PANEL_UPDATE, address >> 1, data);
		return;
	case PIC_REGIONCONFIG:
		{
		uint32_t i = (address - pic24_map[pic24_index].configaddr) >> 1;
		pic24_conf.config[PIC24_PART_UPPER][i] =
			(data & PIC24_CONFIG_MASK) | PIC24_CONFIG_PENDING;
		}
		return;
	case PIC_REGIONID:
		{
		uint32_t i = (address - pic24_map[pic24_index].fuidaddr) >> 1;
		pic24_conf.fuid[i] = (data & PIC24_CONFIG_MASK) | PIC24_CONFIG_PENDING;
		}
	case PIC_REGIONBOOT: /* IGNORED: COMMAND LINE ONLY */
		return;
	}
	if (p.f) fprintf(p.f, "%s: warning: region unsupported [%d]\n",
		__func__, region);
}

/*
 * INIT REGION FOR VERIFY
 *
 *  RETURN REGION IF VERIFY SUPPORTED
 */
static inline uint32_t
pic24_init_verifyregion(uint32_t region)
{
	if (region != PIC_REGIONNOTSUP) {
		return region;
	}
	if (p.f) fprintf(p.f, "%s: warning: region unsupported [%d]\n",
		__func__, region);
	return PIC_REGIONNOTSUP;
}

/*
 * GET VERIFY DATA FOR REGION
 */
static inline uint32_t
pic24_verifyregion(uint32_t address, uint32_t region, uint16_t index, uint32_t wdata)
{
	switch (region) {
	case PIC_REGIONNOTSUP:
		if (p.f) fprintf(p.f, "%s: warning: region unsupported [%d]\n",
			__func__, region);
	case PIC_REGIONCONFIG: /* UNSUPPORTED: BIT MASKS NEEDED  */
		/* FIXME */
	case PIC_REGIONBOOT:   /* IGNORED:     COMMAND LINE ONLY */
		return wdata;
	}
	if (index == 0) {
		pic24_set_read_pointer(address);
		pic24_goto200();
	}

	return pic24_table_read24_post_increment();
}

/*****************************************************************************
 *
 * Program & verify
 *
 *****************************************************************************/

/*
 * BEGIN PROGRAMMING
 */
void
pic24_program_begin(void)
{
	pic24_program_verify();
	pic24_write_program_init();
}

/*
 * PROGRAM DATA
 */
uint32_t
pic24_program_data(uint32_t current_region, pic_data *pdata)
{
	uint32_t address, new_region, wdata;

	for (uint32_t i = 0; i < pdata->nbytes; i += 4) {
		address = (pdata->address + i) >> 1;
		new_region = pic24_getregion(address);
		if (new_region != current_region) {
			pic_write_panel(PIC_PANEL_END, PIC_VOID, PIC_VOID);
			current_region = pic24_init_writeregion(new_region);
		}
		if (current_region == PIC_REGIONNOTSUP)
			continue;
		wdata = pdata->bytes[i] |
			(pdata->bytes[i + 1] << 8) |
			(pdata->bytes[i + 2] << 16);
		pic24_writeregion(address, current_region, wdata);
	}
	return current_region;
}

/*
 * END PROGRAMMING
 */
void
pic24_program_end(int config)
{
	pic_write_panel(PIC_PANEL_END, PIC_VOID, PIC_VOID);
	pic24_standby();
	if (config)
		pic24_write_config();
}

/*
 * VERIFY DATA
 */
uint32_t
pic24_verify_data(uint32_t current_region, pic_data *pdata, uint32_t *fail)
{
	uint32_t address, new_region, wdata, vdata;

	for (uint32_t i = 0; i < pdata->nbytes; i += 4) {
		address = (pdata->address + i) >> 1;
		new_region = pic24_getregion(address);
		if (new_region != current_region)
			current_region = pic24_init_verifyregion(new_region);
		if (current_region == PIC_REGIONNOTSUP)
			continue;
		wdata = pdata->bytes[i] | pdata->bytes[i + 1] << 8 | pdata->bytes[i + 2] << 16;
		vdata = pic24_verifyregion(address, current_region, i, wdata);
		if (vdata != wdata) {
			if (p.f) fprintf(p.f, "%s: error: read [%06X] expected [%06X] at [%06X]\n",
				__func__, vdata, wdata, address);
			pdata->bytes[i] = vdata;
			pdata->bytes[i + 1] = vdata >> 8;
			pdata->bytes[i + 2] = vdata >> 16;
			(*fail) += 4;
		}
	}
	return current_region;
}

/*
 * VIEW DATA
 */
void
pic24_view_data(pic_data *pdata)
{
	uint32_t wdata;

	printf("[%06X] ", pdata->address >> 1);
	for (uint32_t i = 0; i < pdata->nbytes; i += 4) {
		wdata = pdata->bytes[i] |
			(pdata->bytes[i + 1] << 8) |
			(pdata->bytes[i + 2] << 16);
		printf("%06X ", wdata);
	}
	putchar('\n');
}

/*****************************************************************************
 *
 * Diagnostic functions
 *
 *****************************************************************************/

/*
 * DUMP DEVICE ID DETAILS
 */
void
pic24_dumpdeviceid(void)
{
	if (pic24_has_partition())
		printf("[000000] [PROGRAM]   %04X WORDS\n", pic24_map[pic24_index].flash >> 1);
	else
		printf("[000000] [PROGRAM]   %04X WORDS\n", pic24_map[pic24_index].flash);
	uint32_t eaddr;
	if (pic24_get_data_size(&eaddr)) {
		printf("[%06X] [DATA]      %04X BYTES\n",
			eaddr, pic24_map[pic24_index].eeprom);
	}

	/* dsPIC30F */
	if ((pic24_map[pic24_index].datasheet == DS70102H) ||		/* dsPIC30F      */
		(pic24_map[pic24_index].datasheet == DS70284B)) {	/* dsPIC30F SMPS */
		for (uint32_t i = 0; i < 32; i += 4) {
			printf("[%06X] [UNIT%02X]    %04X %04X %04X %04X\n", pic24_map[pic24_index].fuidaddr + 2 * i, i,
				0xFFFF & pic24_conf.fuid[i + 0], 0xFFFF & pic24_conf.fuid[i + 1],
				0xFFFF & pic24_conf.fuid[i + 2], 0xFFFF & pic24_conf.fuid[i + 3]);
		}
		pic24_dumpconfig(PIC_BRIEF, 0);
		int16_t id  = (pic24_conf.deviceid & dsPIC30F_MASK) >> 6;
		int16_t var = (pic24_conf.deviceid & dsPIC30F_VAR_MASK);
		printf("[%06X] [DEVID]     %04X MASK:%03X VAR:%02X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, id, var, pic24_map[pic24_index].devicename);
		int16_t proc = (pic24_conf.revision & dsPIC30F_PROC_MASK) >> 12;
		int16_t rev  = (pic24_conf.revision & dsPIC30F_REV_MASK) >> 6;
		int16_t dot  = (pic24_conf.revision & dsPIC30F_DOT_MASK);
		printf("[%06X] [DEVREV]    %04X PROC:%X REV:%1X DOT:%1X\n",
			PIC24_DEVREV, pic24_conf.revision, proc, rev, dot);
	}

	/* dsPIC33F/PIC24H */
	else if (pic24_map[pic24_index].datasheet == DS70152H) {
		for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i) {
			printf("[%06X] [CALIB%d]    %06X\n", pic24_map[pic24_index].calibaddr + 2 * i,
				i + 1, pic24_conf.calib[i]);
		}
		pic24_dumpconfig(PIC_BRIEF, 0);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nfuid; ++i) {
			printf("[%06X] [FUID%d]     %02X\n",
				pic24_map[pic24_index].fuidaddr + 2 * i, i, 0xFF & pic24_conf.fuid[i]);
		}
		printf("[%06X] [DEVID]     %04X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, pic24_map[pic24_index].devicename);
		printf("[%06X] [DEVREV]    %04X\n",
			PIC24_DEVREV, pic24_conf.revision);
	}

	/* PIC24F KA KL/KM */
	else if (pic24_map[pic24_index].datasheet == DS39919C ||
		 pic24_map[pic24_index].datasheet == DS30625D) {
		for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i) {
			printf("[%06X] [CALIB%d]    %06X\n", pic24_map[pic24_index].calibaddr + 2 * i,
				i + 1, pic24_conf.calib[i]);
		}
		pic24_dumpconfig(PIC_BRIEF, 0);
		int16_t family = (pic24_conf.deviceid & 0xFF00) >> 8;
		int16_t device = (pic24_conf.deviceid & 0x00FF);
		printf("[%06X] [DEVID]     %04X FAM:%02X DEV:%02X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, family, device, pic24_map[pic24_index].devicename);
		int16_t rev = (pic24_conf.revision & 0x0F);
		printf("[%06X] [DEVREV]    %04X REV:%1X\n",
			PIC24_DEVREV, pic24_conf.revision, rev);
	}

	/* dsPIC33F/PIC24FJXXMC with volatile config */
	else if (pic24_map[pic24_index].datasheet == DS70659C ||
		pic24_map[pic24_index].datasheet == DS75012B) {
		pic24_dumpconfig(PIC_BRIEF, 0);
		printf("[%06X] [DEVID]     %04X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, pic24_map[pic24_index].devicename);
		printf("[%06X] [DEVREV]    %04X\n",
			PIC24_DEVREV, pic24_conf.revision);
	}

	/* PIC24F with 2 or 3 volatile config words */
	else if (pic24_map[pic24_index].datasheet == DS39768D ||
		pic24_map[pic24_index].datasheet == DS30009907C) {
		pic24_dumpconfig(PIC_BRIEF, 0);
		for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i) {
			printf("[%06X] [CALIB%d]    %06X\n", pic24_map[pic24_index].calibaddr + 2 * i,
				i + 1, pic24_conf.calib[i]);
		}
		/* FAMID<13:6> DEV<5:0> */
		int16_t family = (pic24_conf.deviceid & PIC24FJ_FAMILY_MASK) >> 6;
		int16_t device = (pic24_conf.deviceid & PIC24FJ_DEV_MASK);
		printf("[%06X] [DEVID]     %04X FAM:%02X DEV:%02X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, family, device, pic24_map[pic24_index].devicename);
		/* MAJRV<8:6> DOT<2:0> */
		int16_t major = (pic24_conf.revision & PIC24FJ_MAJRV_MASK) >> 6;
		int16_t dot   = (pic24_conf.revision & PIC24FJ_DOT_MASK);
		printf("[%06X] [DEVREV]    %04X MAJOR:%1X DOT:%1X\n",
			PIC24_DEVREV, pic24_conf.revision, major, dot);
	}

	/* PIC24F with 4 volatile config words */
	else if (pic24_map[pic24_index].datasheet == DS39934B ||
		pic24_map[pic24_index].datasheet == DS30000510E ||
		pic24_map[pic24_index].datasheet == DS39970E) {
		pic24_dumpconfig(PIC_BRIEF, 0);
		for (uint32_t i = 0; i < pic24_map[pic24_index].ncalib; ++i) {
			printf("[%06X] [CALIB%d]    %06X\n", pic24_map[pic24_index].calibaddr + 2 * i,
				i + 1, pic24_conf.calib[i]);
		}
		/* FAMID<15:8> DEV<7:0> */
		int16_t family = (pic24_conf.deviceid & 0xFF00) >> 8;
		int16_t device = (pic24_conf.deviceid & 0x00FF);
		printf("[%06X] [DEVID]     %04X FAM:%02X DEV:%02X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, family, device, pic24_map[pic24_index].devicename);
		/* REV<3:0> */
		int16_t rev = (pic24_conf.revision & 0x0F);
		printf("[%06X] [DEVREV]    %04X REV:%1X\n",
			PIC24_DEVREV, pic24_conf.revision, rev);
	}

	/* dsPIC33EP/PIC24EP with volatile config */
	else if (pic24_map[pic24_index].datasheet == DS70663D) {
		pic24_dumpconfig(PIC_BRIEF, 0);
		for (uint32_t i = 0; i < pic24_map[pic24_index].nfuid; ++i) {
			printf("[%06X] [FUID%d]     %04X\n",
				pic24_map[pic24_index].fuidaddr + 2 * i, i, 0xFFFF & pic24_conf.fuid[i]);
		}
		printf("[%06X] [DEVID]     %04X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, pic24_map[pic24_index].devicename);
		printf("[%06X] [DEVREV]    %04X\n",
			PIC24_DEVREV, pic24_conf.revision);
	}

	/* dsPIC/PIC24 */
	else {
		pic24_dumpconfig(PIC_BRIEF, 0);
		if (pic24_has_partition()) {
			printf("[%06X] [PROGRAM]   %04X WORDS\n", PIC24_CODE_UPPER,
				pic24_map[pic24_index].flash >> 1);
			pic24_dumpconfig(PIC_BRIEF, 1);
		}
		printf("[%06X] [DEVID]     %04X %s\n",
			PIC24_DEVID, pic24_conf.deviceid, pic24_map[pic24_index].devicename);
		printf("[%06X] [DEVREV]    %04X\n",
			PIC24_DEVREV, pic24_conf.revision);
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig_ga4(uint32_t configaddr, uint32_t configval[])
{
	char config[25][11] = {
		/* 0x00XF80 */ "[FSEC]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x0XXF84 */ "[UNKNOWN]",
		/* 0x0XXF86 */ "[UNKNOWN]",
		/* 0x0XXF88 */ "[UNKNOWN]",
		/* 0x0XXF8A */ "[UNKNOWN]",
		/* 0x0XXF8C */ "[UNKNOWN]",
		/* 0x0XXF8E */ "[UNKNOWN]",
		/* 0x00XF90 */ "[FBSLIM]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x00XF94 */ "[FSIGN]",
		/* 0x0XXF96 */ "[UNKNOWN]",
		/* 0x00XF98 */ "[FOSCSEL]",
		/* 0x0XXF9A */ "[UNKNOWN]",
		/* 0x00XF9C */ "[FOSC]",
		/* 0x0XXF9E */ "[UNKNOWN]",
		/* 0x00XFA0 */ "[FWDT]",
		/* 0x0XXFA2 */ "[UNKNOWN]",
		/* 0x00XFA4 */ "[FPOR]",
		/* 0x0XXFA6 */ "[UNKNOWN]",
		/* 0x00XFA8 */ "[FICD]",
		/* 0x0XXFAA */ "[UNKNOWN]",
		/* 0x00XFAC */ "[FDS]",
		/* 0x0XXFAE */ "[UNKNOWN]",
		/* 0x00XFB0 */ "[FDEVOPT1]"
	};
	for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
		if (strcmp(config[i], "[UNKNOWN]"))
			printf("[%06X] %-11s %04X\n", configaddr + 2 * i,
				config[i], 0xFFFF & configval[i]);
	}
	if (pic24_supports_partition()) {
		printf("[%06X] [FBTSEQ]    %06X\n", PIC24_FBTSEQ | configaddr,
			configval[PIC24_FBTSEQ]);
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig_ga6(uint32_t configaddr, uint32_t configval[])
{
	char config[23][11] = {
		/* 0x00XF00 */ "[FSEC]",
		/* 0x0XXF02 */ "[UNKNOWN]",
		/* 0x0XXF04 */ "[UNKNOWN]",
		/* 0x0XXF06 */ "[UNKNOWN]",
		/* 0x0XXF08 */ "[UNKNOWN]",
		/* 0x0XXF0A */ "[UNKNOWN]",
		/* 0x0XXF0C */ "[UNKNOWN]",
		/* 0x0XXF0E */ "[UNKNOWN]",
		/* 0x00XF10 */ "[FBSLIM]",
		/* 0x0XXF12 */ "[UNKNOWN]",
		/* 0x00XF14 */ "[FSIGN]",
		/* 0x0XXF16 */ "[UNKNOWN]",
		/* 0x00XF18 */ "[FOSCSEL]",
		/* 0x0XXF1A */ "[UNKNOWN]",
		/* 0x00XF1C */ "[FOSC]",
		/* 0x0XXF1E */ "[UNKNOWN]",
		/* 0x00XF20 */ "[FWDT]",
		/* 0x0XXF22 */ "[UNKNOWN]",
		/* 0x00XF24 */ "[FPOR]",
		/* 0x0XXF26 */ "[UNKNOWN]",
		/* 0x00XF28 */ "[FICD]",
		/* 0x0XXF2A */ "[UNKNOWN]",
		/* 0x00XF2C */ "[FDEVOPT1]"
	};
	for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
		if (strcmp(config[i], "[UNKNOWN]"))
			printf("[%06X] %-11s %04X\n", configaddr + 2 * i,
				config[i], 0xFFFF & configval[i]);
	}
	if (pic24_supports_partition()) {
		printf("[%06X] [FBTSEQ]    %06X\n", PIC24_FBTSEQ | configaddr,
			configval[PIC24_FBTSEQ]);
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig_gs(uint32_t configaddr, uint32_t configval[])
{
	char config[25][11] = {
		/* 0x00XF80 */ "[FSEC]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x0XXF84 */ "[UNKNOWN]",
		/* 0x0XXF86 */ "[UNKNOWN]",
		/* 0x0XXF88 */ "[UNKNOWN]",
		/* 0x0XXF8A */ "[UNKNOWN]",
		/* 0x0XXF8C */ "[UNKNOWN]",
		/* 0x0XXF8E */ "[UNKNOWN]",
		/* 0x00XF90 */ "[FBSLIM]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x00XF94 */ "[FSIGN]",
		/* 0x0XXF96 */ "[UNKNOWN]",
		/* 0x00XF98 */ "[FOSCSEL]",
		/* 0x0XXF9A */ "[UNKNOWN]",
		/* 0x00XF9C */ "[FOSC]",
		/* 0x0XXF9E */ "[UNKNOWN]",
		/* 0x00XFA0 */ "[FWDT]",
		/* 0x0XXFA2 */ "[UNKNOWN]",
		/* 0x00XFA4 */ "[FPOR]",
		/* 0x0XXFA6 */ "[UNKNOWN]",
		/* 0x00XFA8 */ "[FICD]",
		/* 0x0XXFAA */ "[UNKNOWN]",
		/* 0x00XFAC */ "[FDEVOPT]",
		/* 0x0XXFAE */ "[UNKNOWN]",
		/* 0x00XFB0 */ "[FALTREG]"
	};
	for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
		if (strcmp(config[i], "[UNKNOWN]"))
			printf("[%06X] %-11s %04X\n", configaddr + 2 * i,
				config[i], 0xFFFF & configval[i]);
	}
	if (pic24_supports_partition()) {
		printf("[%06X] [FBTSEQ]    %06X\n", PIC24_FBTSEQ | configaddr,
			configval[PIC24_FBTSEQ]);
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig_gs2(uint32_t configaddr, uint32_t configval[])
{
	char config[25][11] = {
		/* 0x00XF80 */ "[UNKNOWN]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x0XXF84 */ "[UNKNOWN]",
		/* 0x0XXF86 */ "[UNKNOWN]",
		/* 0x0XXF88 */ "[UNKNOWN]",
		/* 0x0XXF8A */ "[UNKNOWN]",
		/* 0x0XXF8C */ "[FSEC]",
		/* 0x0XXF8E */ "[UNKNOWN]",
		/* 0x00XF90 */ "[FBSLIM]",
		/* 0x0XXF92 */ "[UNKNOWN]",
		/* 0x00XF94 */ "[FSIGN]",
		/* 0x0XXF96 */ "[UNKNOWN]",
		/* 0x00XF98 */ "[FOSCSEL]",
		/* 0x0XXF9A */ "[UNKNOWN]",
		/* 0x00XF9C */ "[FOSC]",
		/* 0x0XXF9E */ "[UNKNOWN]",
		/* 0x00XFA0 */ "[FWDT]",
		/* 0x0XXFA2 */ "[UNKNOWN]",
		/* 0x00XFA4 */ "[FPOR]",
		/* 0x0XXFA6 */ "[UNKNOWN]",
		/* 0x00XFA8 */ "[FICD]",
		/* 0x0XXFAA */ "[UNKNOWN]",
		/* 0x00XFAC */ "[FDEVOPT]",
		/* 0x0XXFAE */ "[UNKNOWN]",
		/* 0x00XFB0 */ "[FALTREG]"
	};
	for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
		if (strcmp(config[i], "[UNKNOWN]"))
			printf("[%06X] %-11s %04X\n", configaddr + 2 * i,
				config[i], 0xFFFF & configval[i]);
	}
	if (pic24_supports_partition()) {
		printf("[%06X] [FBTSEQ]    %06X\n", PIC24_FBTSEQ | configaddr,
			configval[PIC24_FBTSEQ]);
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig_switch(uint32_t configaddr, uint32_t configval[])
{
	switch (pic24_map[pic24_index].datasheet) {
	case DS30010073A: /* PIC24FJ64GA412    */
		pic24_dumpconfig_ga4(configaddr, configval);
		break;
	case DS30010057F: /* PIC24FJ128GA610   */
	case DS30010102A: /* PIC24FJ64GA702    */
		pic24_dumpconfig_ga6(configaddr, configval);
		break;
	case DS70005160C: /* dsPIC33EP64GS502  */
	case DS70005256A: /* dsPIC33EP128GS702 */
		pic24_dumpconfig_gs(configaddr, configval);
		break;
	case DS70005192B: /* dsPIC33EP16GS202 */
		pic24_dumpconfig_gs2(configaddr, configval);
		break;
	default:printf("[UNIMPLEMENTED]\n");
		break;
	}
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE
 */
void
pic24_dumpconfig(uint32_t mode, uint32_t partition)
{
	/* dsPIC30F */
	if (pic24_map[pic24_index].datasheet == DS70102H) {
		char config[7][10] = {
			/* 0xF80000 */ "[FOSC]",
			/* 0xF80002 */ "[FWDT]",
			/* 0xF80004 */ "[FBORPOR]",
			/* 0xF80006 */ "[FBS]",
			/* 0xF80008 */ "[FSS]",
			/* 0xF8000A */ "[FGS]",
			/* 0xF8000C */ "[FICD]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			printf("[%06X] %-11s %04X\n", pic24_map[pic24_index].configaddr + 2 * i,
				config[i], 0xFFFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/* dsPIC30F SMPS */
	else if (pic24_map[pic24_index].datasheet == DS70284B) {
		char config[8][11] = {
			/* 0xF80000 */ "[FBS]",
			/* 0xF80002 */ "[UNKNOWN]",
			/* 0xF80004 */ "[FGS]",
			/* 0xF80006 */ "[FOSCSEL]",
			/* 0xF80008 */ "[FOSC]",
			/* 0xF8000A */ "[FWDT]",
			/* 0xF8000C */ "[FPOR]",
			/* 0xF8000E */ "[FICD]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			if (strcmp(config[i], "[UNKNOWN]"))
				printf("[%06X] %-11s %02X\n", pic24_map[pic24_index].configaddr + 2 * i,
					config[i], 0xFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/* dsPIC33F/PIC24H */
	else if (pic24_map[pic24_index].datasheet == DS70152H) {
		char config[9][10] = {
			/* 0xF80000 */ "[FBS]",
			/* 0xF80002 */ "[FSS]",
			/* 0xF80004 */ "[FGS]",
			/* 0xF80006 */ "[FOSCSEL]",
			/* 0xF80008 */ "[FOSC]",
			/* 0xF8000A */ "[FWDT]",
			/* 0xF8000C */ "[FPOR]",
			/* 0xF8000E */ "[FICD]",
			/* 0xF80010 */ "[FCMP]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			printf("[%06X] %-11s %02X\n", pic24_map[pic24_index].configaddr + 2 * i,
				config[i], 0xFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/* PIC24F KA KL/KM */
	else if (pic24_map[pic24_index].datasheet == DS39919C ||
		 pic24_map[pic24_index].datasheet == DS30625D) {
		char config[9][11] = {
			/* 0xF80000 */ "[FBS]",
			/* 0xF80002 */ "[UNKNOWN]",
			/* 0xF80004 */ "[FGS]",
			/* 0xF80006 */ "[FOSCSEL]",
			/* 0xF80008 */ "[FOSC]",
			/* 0xF8000A */ "[FWDT]",
			/* 0xF8000C */ "[FPOR]",
			/* 0xF8000E */ "[FICD]",
			/* 0xF80010 */ "[FDS]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			if (strcmp(config[i], "[UNKNOWN]"))
				printf("[%06X] %-11s %02X\n", pic24_map[pic24_index].configaddr + 2 * i,
					config[i], 0xFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/* dsPIC33F/PIC24FJXXMC with volatile config */
	else if (pic24_map[pic24_index].datasheet == DS70659C ||
		pic24_map[pic24_index].datasheet == DS75012B) {
		if (pic24_map[pic24_index].nconfig == 2) {
			printf("[%06X] [CONFIG2]   %04X\n",
				pic24_map[pic24_index].configaddr + 0, 0xFFFF & pic24_conf.config[PIC24_CONFIG][0]);
			printf("[%06X] [CONFIG1]   %04X\n",
				pic24_map[pic24_index].configaddr + 2, 0xFFFF & pic24_conf.config[PIC24_CONFIG][1]);
		} else {
			char config[8][11] = {
				/* 0x00XXF0 */ "[FICD]",
				/* 0x00XXF2 */ "[UNKNOWN]",
				/* 0x00XXF4 */ "[FWDT]",
				/* 0x00XXF6 */ "[FOSC]",
				/* 0x00XXF8 */ "[FOSCSEL]",
				/* 0x00XXFA */ "[FGS]",
				/* 0x00XXFC */ "[UNKNOWN]",
				/* 0x00XXFE */ "[UNKNOWN]"
			};
			for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
				if (strcmp(config[i], "[UNKNOWN]"))
					printf("[%06X] %-11s %02X\n", pic24_map[pic24_index].configaddr + 2 * i,
						config[i], 0xFF & pic24_conf.config[PIC24_CONFIG][i]);
			}
		}
	}

	/* PIC24F with 2 volatile config words */
	else if (pic24_map[pic24_index].datasheet == DS39768D) {
		printf("[%06X] [CONFIG2]   %04X\n",
			pic24_map[pic24_index].configaddr + 0, 0xFFFF & pic24_conf.config[PIC24_CONFIG][0]);
		printf("[%06X] [CONFIG1]   %04X\n",
			pic24_map[pic24_index].configaddr + 2, 0xFFFF & pic24_conf.config[PIC24_CONFIG][1]);
	}

	/* PIC24F with 3 volatile config words */
	else if (pic24_map[pic24_index].datasheet == DS30009907C) {
		printf("[%06X] [CONFIG3]   %04X\n",
			pic24_map[pic24_index].configaddr + 0, 0xFFFF & pic24_conf.config[PIC24_CONFIG][0]);
		printf("[%06X] [CONFIG2]   %04X\n",
			pic24_map[pic24_index].configaddr + 2, 0xFFFF & pic24_conf.config[PIC24_CONFIG][1]);
		printf("[%06X] [CONFIG1]   %04X\n",
			pic24_map[pic24_index].configaddr + 4, 0xFFFF & pic24_conf.config[PIC24_CONFIG][2]);
	}

	/* PIC24F with 4 volatile config words */
	else if (pic24_map[pic24_index].datasheet == DS39934B ||
		pic24_map[pic24_index].datasheet == DS30000510E ||
		pic24_map[pic24_index].datasheet == DS39970E) {
		printf("[%06X] [CONFIG4]   %04X\n",
			pic24_map[pic24_index].configaddr + 0, 0xFFFF & pic24_conf.config[PIC24_CONFIG][0]);
		printf("[%06X] [CONFIG3]   %04X\n",
			pic24_map[pic24_index].configaddr + 2, 0xFFFF & pic24_conf.config[PIC24_CONFIG][1]);
		printf("[%06X] [CONFIG2]   %04X\n",
			pic24_map[pic24_index].configaddr + 4, 0xFFFF & pic24_conf.config[PIC24_CONFIG][2]);
		printf("[%06X] [CONFIG1]   %04X\n",
			pic24_map[pic24_index].configaddr + 6, 0xFFFF & pic24_conf.config[PIC24_CONFIG][3]);
	}

	/* dsPIC33EP/PIC24EP with volatile config */
	else if (pic24_map[pic24_index].datasheet == DS70663D) {
		char config[10][11] = {
			/* 0x00XXEC */ "[UNKNOWN]",
			/* 0x00XXEE */ "[UNKNOWN]",
			/* 0x00XXF0 */ "[FICD]",
			/* 0x00XXF2 */ "[FPOR]",
			/* 0x00XXF4 */ "[FWDT]",
			/* 0x00XXF6 */ "[FOSC]",
			/* 0x00XXF8 */ "[FOSCSEL]",
			/* 0x00XXFA */ "[FGS]",
			/* 0x00XXFC */ "[UNKNOWN]",
			/* 0x00XXFE */ "[UNKNOWN]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			if (strcmp(config[i], "[UNKNOWN]"))
				printf("[%06X] %-11s %02X\n", pic24_map[pic24_index].configaddr + 2 * i,
					config[i], 0xFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/* dsPIC33EV with volatile config */
	else if (pic24_map[pic24_index].datasheet == DS70005137F) {
		char config[35][12] = {
			/* 0x0XXX80 */ "[FSEC]",
			/* 0x0XXX82 */ "[UNKNOWN]",
			/* 0x0XXX84 */ "[UNKNOWN]",
			/* 0x0XXX86 */ "[UNKNOWN]",
			/* 0x0XXX88 */ "[UNKNOWN]",
			/* 0x0XXX8A */ "[UNKNOWN]",
			/* 0x0XXX8C */ "[UNKNOWN]",
			/* 0x0XXX8E */ "[UNKNOWN]",
			/* 0x0XXX90 */ "[FSBLIM]",
			/* 0x0XXX92 */ "[UNKNOWN]",
			/* 0x0XXX94 */ "[RESERVED]",
			/* 0x0XXX96 */ "[UNKNOWN]",
			/* 0x0XXX98 */ "[FOSCSEL]",
			/* 0x0XXX9A */ "[UNKNOWN]",
			/* 0x0XXX9C */ "[FOSC]",
			/* 0x0XXX9E */ "[UNKNOWN]",
			/* 0x0XXXA0 */ "[FWDT]",
			/* 0x0XXXA2 */ "[UNKNOWN]",
			/* 0x0XXXA4 */ "[FPOR]",
			/* 0x0XXXA6 */ "[UNKNOWN]",
			/* 0x0XXXA8 */ "[FICD]",
			/* 0x0XXXAA */ "[UNKNOWN]",
			/* 0x0XXXAC */ "[FDMTINTVL]",
			/* 0x0XXXAE */ "[UNKNOWN]",
			/* 0x0XXXB0 */ "[FDMTINTVH]",
			/* 0x0XXXB2 */ "[UNKNOWN]",
			/* 0x0XXXB4 */ "[FDMTCNTL]",
			/* 0x0XXXB2 */ "[UNKNOWN]",
			/* 0x0XXXB8 */ "[FDMTCNTH]",
			/* 0x0XXXBA */ "[UNKNOWN]",
			/* 0x0XXXBC */ "[FDMT]",
			/* 0x0XXXBE */ "[UNKNOWN]",
			/* 0x0XXXC0 */ "[FDEVOPT]",
			/* 0x0XXXC2 */ "[UNKNOWN]",
			/* 0x0XXXC4 */ "[FALTREG]"
		};
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			if (strcmp(config[i], "[UNKNOWN]"))
				printf("[%06X] %-11s %04X\n", pic24_map[pic24_index].configaddr + 2 * i,
					config[i], 0xFFFF & pic24_conf.config[PIC24_CONFIG][i]);
		}
	}

	/*
	 * dsPIC33EPXXGS5XX     with volatile config
	 * PIC24FJXXGA4/GB4     with volatile config
	 * PIC24FJXXGA6/GB6/GA7 with volatile config
	 */
	else if ((pic24_map[pic24_index].datasheet == DS70005160C) || /* dsPIC33EP64GS502  */
		(pic24_map[pic24_index].datasheet == DS70005192B)  || /* dsPIC33EP16GS202  */
		(pic24_map[pic24_index].datasheet == DS70005256A)  || /* dsPIC33EP128GS702 */
		(pic24_map[pic24_index].datasheet == DS30010073A)  || /* PIC24FJ64GA412    */
		(pic24_map[pic24_index].datasheet == DS30010057F)  || /* PIC24FJ128GB606   */
		(pic24_map[pic24_index].datasheet == DS30010102A)) {  /* PIC24FJ64GA702    */
		if (pic24_has_partition()) {
			if (partition == 0) {
				pic24_dumpconfig_switch(pic24_map[pic24_index].configlower,
					pic24_conf.config[PIC24_PART_LOWER]);
			} else {
				pic24_dumpconfig_switch(pic24_map[pic24_index].configlower | PIC24_CODE_UPPER,
					pic24_conf.config[PIC24_PART_UPPER]);
				printf("[%06X] [FBOOT]     %06X\n",
					pic24_map[pic24_index].fbootaddr, pic24_conf.fboot);
			}
		} else {
			pic24_dumpconfig_switch(pic24_map[pic24_index].configaddr,
				pic24_conf.config[PIC24_CONFIG]);
			if (pic24_supports_partition()) {
				printf("[%06X] [FBOOT]     %06X\n",
					pic24_map[pic24_index].fbootaddr, pic24_conf.fboot);
			}
		}
	}

	else {
		printf("[UNIMPLEMENTED]\n");
	}

#ifdef VERBOSE
	if (mode == PIC_VERBOSE)
		pic24_dumpconfig_verbose();
#endif
}

/*
 * DUMP VERBOSE CONFIG WORD DETAILS FOR DEVICE
 */
#ifdef VERBOSE
void
pic24_dumpconfig_verbose(void)
{
	/* UNIMPLEMENTED */
}
#endif /* VERBOSE */

/*
 * DUMP HEX FLASH WORDS
 */
void
pic24_dumphexcode(uint32_t address, uint32_t size, uint32_t *data)
{
	uint32_t i, j, nlines = 0;

	for (i = 0; i < size; address += 16, i += 8) {
#ifndef DEBUG
		if (pic_mtcode(PIC24_MASK, 8, &data[i]))
			continue;
#endif
		printf("[%06X] ", address);
		for (j = 0; j < 8; ++j)
			printf("%06X ", data[i + j] & PIC24_MASK);
		for (j = 0; j < 8; ++j) {
			putchar(PIC_CHAR(0xFF & data[i + j]));
			putchar(PIC_CHAR(0xFF & (data[i + j] >> 8)));
			putchar(PIC_CHAR(0xFF & (data[i + j] >> 16)));
		}
		putchar('\n');
		nlines++;
	}
	if (!nlines)
		printf("%s: information: flash empty\n", __func__);
}

/*
 * DUMP INHX32 FLASH WORDS
 */
void
pic24_dumpinhxcode(uint32_t address, uint32_t size, uint32_t *data)
{
	uint32_t i, j;

	/* PIC24: Extended address */
	pic_dumpaddr(address << 1, 1);

	for (i = 0; i < size; address += 8, i += 4) {
		if (pic_mtcode(PIC24_MASK, 4, &data[i]))
			continue;
		/* PIC24: Extended address */
		pic_dumpaddr(address << 1, 0);

		uint8_t cc, hb, mb, lb;
		mb = address >> 7;
		lb = address << 1;
		printf(":%02X%02X%02X00", 16, mb, lb);
		cc = 16 + mb + lb + 0x00;
		for (j = 0; j < 4; ++j) {
			lb = data[i + j];
			mb = data[i + j] >> 8;
			hb = data[i + j] >> 16;
			printf("%02X%02X%02X00", lb, mb, hb);
			cc = cc + lb + mb + hb + 0x00;
		}
		printf("%02X\n", (0x0100 - cc) & 0xFF);
	}
}

/*
 * DUMP HEX DATA EEPROM WORDS
 */
void
pic24_dumphexdata(uint32_t address, uint32_t size, uint16_t *data)
{
	uint32_t i, j, nlines = 0;

	for (i = 0; i < (size / 2); address += 16, i += 8) {
		if (pic_mtdata(0xFFFF, 8, &data[i]))
			continue;
		printf("[%06X] ", address);
		for (j = 0; j < 8; ++j)
			printf("%04X ", data[i + j]);
		for (j = 0; j < 8; ++j) {
			putchar(PIC_CHAR(0xFF & data[i + j]));
			putchar(PIC_CHAR(0xFF & (data[i + j] >> 8)));
		}
		putchar('\n');
		nlines++;
	}
	if (!nlines)
		printf("%s: information: data empty\n", __func__);
}

/*
 * DUMP INHX32 DATA EEPROM WORDS
 */
void
pic24_dumpinhxdata(uint32_t address, uint32_t size, uint16_t *data)
{
	uint32_t i, j;

	/* PIC24: Extended address */
	pic_dumpaddr(address << 1, 1);

	for (i = 0; i < (size / 2); address += 8, i += 4) {
		if (pic_mtdata(0xFFFF, 4, &data[i]))
			continue;
		/* PIC24: Extended address */
		pic_dumpaddr(address << 1, 0);

		uint8_t cc, mb, lb;
		mb = address >> 7;
		lb = address << 1;
		printf(":%02X%02X%02X00", 16, mb, lb);
		cc = 16 + mb + lb + 0x00;
		for (j = 0; j < 4; ++j) {
			lb = data[i + j];
			mb = data[i + j] >> 8;
			printf("%02X%02X0000", lb, mb);
			cc = cc + lb + mb + 0x00 + 0x00;
		}
		printf("%02X\n", (0x0100 - cc) & 0xFF);
	}
}

/*
 * DUMP INHX32 DEVICE CONFIGURATION
 */
void
pic24_dumpdevice(void)
{
	if (pic24_has_fuid()) {
		/* PIC24: Extended address */
		pic_dumpaddr(pic24_map[pic24_index].fuidaddr << 1, 1);

		/* FUID */
		for (uint32_t i = 0; i < pic24_map[pic24_index].nfuid; ++i) {
			pic_dumpword32(pic24_map[pic24_index].fuidaddr + 2 * i,
				pic24_conf.fuid[i]);
		}
	}
	if (pic24_has_config()) {
		/* PIC24: Extended address */
		pic_dumpaddr(pic24_map[pic24_index].configaddr << 1, 1);

		/* CONFIG */
		for (uint32_t i = 0; i < pic24_map[pic24_index].nconfig; ++i) {
			pic_dumpword32(pic24_map[pic24_index].configaddr + 2 * i,
				pic24_conf.config[PIC24_PART_UPPER][i]);
		}
	}
}
